"""
Base classes for image transformations.
"""

import abc
from typing import ClassVar
from typing_extensions import override

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Float, Inexact

from ...jax_util import FloatLike


class AbstractImageTransform(eqx.Module, strict=True):
    """Base class for computing and applying an `Array` to an image."""

    is_real_space: eqx.AbstractClassVar[bool]

    @abc.abstractmethod
    def __call__(
        self, image: Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"]
    ) -> Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"]:
        raise NotImplementedError

    def __mul__(self, other) -> "AbstractImageTransform":
        if self.is_real_space:
            return _RealProductImageTransform(transform1=self, transform2=other)
        else:
            return _FourierProductImageTransform(transform1=self, transform2=other)

    def __rmul__(self, other) -> "AbstractImageTransform":
        if self.is_real_space:
            return _RealProductImageTransform(transform1=other, transform2=self)
        else:
            return _FourierProductImageTransform(transform1=other, transform2=self)


class _AbstractProductImageTransform(AbstractImageTransform, strict=True):
    """A helper to represent the product of two transforms."""

    transform1: AbstractImageTransform
    transform2: AbstractImageTransform

    def __check_init__(self):
        if self.transform1.is_real_space != self.transform2.is_real_space:
            raise TypeError(
                "Cannot compose two `AbstractImageTransforms` "
                " with unequal attributes `is_real_space`. Type "
                f"{self.transform1.__class__.__name__} had "
                f"`is_real_space = {self.transform1.is_real_space}`, "
                f"but {self.transform2.__class__.__name__} had "
                f"`is_real_space = {self.transform2.is_real_space}`."
            )

    @override
    def __call__(
        self, image: Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"]
    ) -> Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"]:
        return self.transform1(self.transform2(image))

    def __repr__(self):
        return f"{repr(self.transform1)} * {repr(self.transform2)}"


class _RealProductImageTransform(_AbstractProductImageTransform):
    is_real_space: ClassVar[bool] = True


class _FourierProductImageTransform(_AbstractProductImageTransform):
    is_real_space: ClassVar[bool] = False


class ScaleImage(AbstractImageTransform, strict=True):
    scale: Float[Array, ""]
    offset: Float[Array, ""]

    is_real_space: ClassVar[bool] = True

    def __init__(self, scale: FloatLike = 1.0, offset: FloatLike = 0.0):
        self.scale = jnp.asarray(scale, dtype=float)
        self.offset = jnp.asarray(offset, dtype=float)

    @override
    def __call__(
        self, image: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return self.scale * image + self.offset



"""
Image formation models.
"""

from abc import abstractmethod
from typing import Literal
from typing_extensions import override

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import Array, Bool, Complex, Float, PRNGKeyArray

from ..jax_util import NDArrayLike
from ..ndimage import (
    AbstractFilter,
    AbstractImageTransform,
    AbstractMask,
    compute_edge_value,
    crop_to_shape,
    irfftn,
    rfftn,
)
from ._detector import AbstractDetector
from ._image_config import AbstractImageConfig, DoseImageConfig
from ._pose import AbstractPose
from ._scattering_theory import AbstractScatteringTheory
from ._transfer_theory import ContrastTransferTheory
from ._volume import (
    AbstractAtomVolume,
    AbstractVolumeIntegrator,
    AbstractVolumeParametrization,
    AbstractVolumeRepresentation,
    AutoVolumeProjection,
)


RealImageArray = Float[Array, "{self.image_config.y_dim} {self.image_config.x_dim}"]
FourierImageArray = Complex[
    Array, "{self.image_config.y_dim} {self.image_config.x_dim//2+1}"
]
PaddedRealImageArray = Float[
    Array,
    "{self.image_config.padded_y_dim} {self.image_config.padded_x_dim}",
]
PaddedFourierImageArray = Complex[
    Array,
    "{self.image_config.padded_y_dim} {self.image_config.padded_x_dim//2+1}",
]

ImageArray = RealImageArray | FourierImageArray
PaddedImageArray = PaddedRealImageArray | PaddedFourierImageArray


class AbstractImageModel(eqx.Module, strict=True):
    """Base class for an image formation model.

    Call an `AbstractImageModel`'s `simulate` routine.
    """

    image_config: eqx.AbstractVar[AbstractImageConfig]
    pose: eqx.AbstractVar[AbstractPose]

    image_transform: eqx.AbstractVar[AbstractImageTransform | None]

    signal_region: eqx.AbstractVar[Bool[Array, "_ _"] | None]
    normalizes_signal: eqx.AbstractVar[bool]
    signal_centering: eqx.AbstractVar[Literal["bg", "mean"]]
    translate_mode: eqx.AbstractVar[Literal["fft", "atom", "none"]]

    def __check_init__(self):
        if self.signal_centering not in ["bg", "mean"]:
            raise ValueError(
                "Found invalid value for "
                f"`{self.__class__.__name__}(..., signal_centering=...)`. "
                "Values 'bg' and 'mean' are "
                f"supported, but got {self.signal_centering}."
            )
        if self.translate_mode not in ["fft", "atom", "none"]:
            raise ValueError(
                "Found invalid value for "
                f"`{self.__class__.__name__}(..., translate_mode=...)`. "
                "Values 'fft', 'atom', and 'none' are "
                f"supported, but got {self.translate_mode}."
            )

    @abstractmethod
    def raw_simulate(
        self,
        rng_key: PRNGKeyArray | None = None,
        *,
        outputs_real_space: bool = True,
    ) -> Array:
        """Render an image without postprocessing."""
        raise NotImplementedError

    def simulate(
        self,
        rng_key: PRNGKeyArray | None = None,
        *,
        outputs_real_space: bool = True,
        mask: AbstractMask | None = None,
        filter: AbstractFilter | None = None,
    ) -> Array:
        """Render an image.

        **Arguments:**

        - `rng_key`:
            The random number generator key. If not passed, render an image
            with no stochasticity.
        - `outputs_real_space`:
            If `True`, return the image in real space.
        - `mask`:
            Optionally apply a mask to the image.
        - `filter`:
            Optionally apply a filter to the image.
        """
        fourier_image = self.raw_simulate(rng_key, outputs_real_space=False)

        return self.postprocess(
            fourier_image,
            outputs_real_space=outputs_real_space,
            mask=mask,
            filter=filter,
        )

    def postprocess(
        self,
        fourier_image: Array,
        *,
        outputs_real_space: bool = True,
        mask: AbstractMask | None = None,
        filter: AbstractFilter | None = None,
    ) -> Array:
        """Return an image postprocessed with filters, cropping, masking,
        and normalization in either real or fourier space.
        """
        image_config = self.image_config
        mask_c, filter_c = self._compose_transform(mask, filter)
        if (
            mask_c is None
            and image_config.padded_shape == image_config.shape
            and not self.normalizes_signal
        ):
            # ... if there are no masks, we don't need to crop, and we are
            # not normalizing, minimize moving back and forth between real
            # and fourier space
            if filter_c is not None:
                fourier_image = filter_c(fourier_image)
            return (
                irfftn(fourier_image, s=image_config.shape)
                if outputs_real_space
                else fourier_image
            )
        else:
            # ... otherwise, apply filter, crop, and mask, again trying to
            # minimize moving back and forth between real and fourier space
            padded_rfft_shape = (
                image_config.padded_shape[0],
                image_config.padded_shape[1] // 2 + 1,
            )
            if filter_c is not None:
                # ... apply the filter
                if filter is not None:
                    if not filter.get().shape == padded_rfft_shape:
                        raise ValueError(
                            "Found that the `filter` was shape "
                            f"{filter.get().shape}, but expected it to be "
                            f"shape {padded_rfft_shape}. You may have passed a "
                            f"filter according to the "
                            f"`{image_config.__class__.__name__}.shape`, "
                            f"when the `{image_config.__class__.__name__}.padded_shape` "
                            "was expected."
                        )
                fourier_image = filter_c(fourier_image)
            padded_image = irfftn(fourier_image, s=image_config.padded_shape)
            if image_config.padded_shape != image_config.shape:
                image = crop_to_shape(padded_image, image_config.shape)
            else:
                image = padded_image
            if self.normalizes_signal:
                if self.signal_centering == "mean":
                    image = self._mean_subtract_normalize(image)
                elif self.signal_centering == "bg":
                    image = self._bg_subtract_normalize(image, padded_image)
            if mask_c is not None:
                image = mask_c(image)
            return image if outputs_real_space else rfftn(image)

    def _phase_shift_translate(self, fourier_image: Array) -> Array:
        phase_shifts = self.pose.compute_translation_operator(
            self.image_config.get_frequency_grid(physical=True, padding=True)
        )
        fourier_image = self.pose.translate_image(
            fourier_image,
            phase_shifts,
            self.image_config.padded_shape,
        )

        return fourier_image

    def _atom_translate(self, volrep: AbstractVolumeRepresentation) -> AbstractAtomVolume:
        if isinstance(volrep, AbstractAtomVolume):
            return volrep.translate_to_pose(self.pose)
        else:
            raise ValueError(
                "Tried to apply translation in `translate_mode = 'atom'`, but "
                "found a volume representation that was not an `AbstractAtomVolume`."
                f"Got a `{volrep.__class__.__name__}` class."
            )

    def _compose_transform(
        self, mask: AbstractMask | None, filter: AbstractFilter | None
    ) -> tuple[AbstractImageTransform | None, AbstractImageTransform | None]:
        if self.image_transform is None:
            return mask, filter
        else:
            if self.image_transform.is_real_space:
                if mask is None:
                    return self.image_transform, filter
                else:
                    return mask * self.image_transform, filter
            else:
                if filter is None:
                    return mask, self.image_transform
                else:
                    return mask, filter * self.image_transform

    def _mean_subtract_normalize(self, image: Array) -> Array:
        mean, std = (
            jnp.mean(image, where=self.signal_region),
            jnp.std(image, where=self.signal_region),
        )
        image = (image - mean) / std

        return image

    def _bg_subtract_normalize(self, image: Array, padded_image: Array) -> Array:
        bg_value, std = (
            compute_edge_value(padded_image),
            jnp.std(image, where=self.signal_region),
        )
        image = (image - bg_value) / std

        return image


class LinearImageModel(AbstractImageModel, strict=True):
    """An simple image model in linear image formation theory."""

    volume: AbstractVolumeParametrization
    pose: AbstractPose
    volume_integrator: AbstractVolumeIntegrator
    transfer_theory: ContrastTransferTheory
    image_config: AbstractImageConfig

    image_transform: AbstractImageTransform | None
    normalizes_signal: bool
    signal_region: Bool[Array, "_ _"] | None
    signal_centering: Literal["bg", "mean"]
    translate_mode: Literal["fft", "atom", "none"]

    def __init__(
        self,
        volume: AbstractVolumeParametrization,
        pose: AbstractPose,
        image_config: AbstractImageConfig,
        transfer_theory: ContrastTransferTheory,
        volume_integrator: AbstractVolumeIntegrator = AutoVolumeProjection(),
        *,
        image_transform: AbstractImageTransform | None = None,
        normalizes_signal: bool = False,
        signal_region: Bool[NDArrayLike, "_ _"] | None = None,
        signal_centering: Literal["bg", "mean"] = "mean",
        translate_mode: Literal["fft", "atom", "none"] = "fft",
    ):
        """**Arguments:**

        - `volume`:
            The parametrization of an imaging volume.
        - `pose`:
            The pose of the volume.
        - `image_config`:
            The configuration of the instrument, such as for the pixel size
            and the wavelength.
        - `volume_integrator`: The method for integrating the volume onto the plane.
        - `transfer_theory`: The contrast transfer theory.
        - `image_transform`:
            A [`cryojax.ndimage.AbstractImageTransform`][] applied to the
            image after simulation.
        - `normalizes_signal`:
            Whether or not to normalize the output of `image_model.simulate()`.
            If `True`, see `signal_centering` for options.
        - `signal_region`:
            A boolean array that is 1 where there is signal,
            and 0 otherwise used to normalize the image.
            Must have shape equal to `AbstractImageConfig.shape`.
        - `signal_centering`:
            How to calculate the offset for normalization when
            `normalizes_signal = True`. Options are
            - 'mean':
                Normalize the image to be mean 0
                within `signal_region`.
            - 'bg':
                Subtract mean value at the image edges.
                This makes the image fade to a background with values
                equal to zero. Requires that `image_config.padded_shape`
                is large enough so that the signal sufficiently decays.
            Ignored if `normalizes_signal = False`.
        - `translate_mode`:
            How to apply in-plane translation to the volume. Options are
            - 'fft':
                Apply phase shifts in the Fourier domain.
            - 'atom':
                Apply translation to atom positions before
                projection. For this method, the
                [`cryojax.simulator.AbstractVolumeParametrization`][]
                must be or return an [`cryojax.simulator.AbstractAtomVolume`][].
            - 'none':
                Do not apply the translation.
        """
        # Simulator components
        self.volume = volume
        self.pose = pose
        self.image_config = image_config
        self.volume_integrator = volume_integrator
        self.transfer_theory = transfer_theory
        # Options
        self.image_transform = image_transform
        self.translate_mode = translate_mode
        self.normalizes_signal = normalizes_signal
        self.signal_centering = signal_centering
        if signal_region is None:
            self.signal_region = None
        else:
            self.signal_region = jnp.asarray(signal_region, dtype=bool)

    @override
    def raw_simulate(
        self,
        rng_key: PRNGKeyArray | None = None,
        *,
        outputs_real_space: bool = True,
    ) -> PaddedFourierImageArray:
        # Get the representation of the volume
        if rng_key is None:
            volume_representation = self.volume.to_representation()
        else:
            this_key, rng_key = jr.split(rng_key)
            volume_representation = self.volume.to_representation(rng_key=this_key)
        # Rotate it to the lab frame
        volume_representation = volume_representation.rotate_to_pose(self.pose)
        # Translate if using atom translations
        if self.translate_mode == "atom":
            volume_representation = self._atom_translate(volume_representation)
        # Compute the projection image
        fourier_image = self.volume_integrator.integrate(
            volume_representation, self.image_config, outputs_real_space=False
        )
        # Compute the image
        fourier_image = self.transfer_theory.propagate_object(  # noqa: E501
            fourier_image,
            self.image_config,
            input_is_ewald_sphere=self.volume_integrator.outputs_ewald_sphere,
            defocus_offset=self.pose.offset_z_in_angstroms,
        )
        # Now for the in-plane translation if using phase shifts
        if self.translate_mode == "fft":
            fourier_image = self._phase_shift_translate(fourier_image)

        return (
            irfftn(fourier_image, s=self.image_config.padded_shape)
            if outputs_real_space
            else fourier_image
        )


class ProjectionImageModel(AbstractImageModel, strict=True):
    """An simple image model for computing a projection."""

    volume: AbstractVolumeParametrization
    pose: AbstractPose
    volume_integrator: AbstractVolumeIntegrator
    image_config: AbstractImageConfig

    image_transform: AbstractImageTransform | None
    normalizes_signal: bool
    signal_region: Bool[Array, "_ _"] | None
    signal_centering: Literal["bg", "mean"]
    translate_mode: Literal["fft", "atom", "none"]

    def __init__(
        self,
        volume: AbstractVolumeParametrization,
        pose: AbstractPose,
        image_config: AbstractImageConfig,
        volume_integrator: AbstractVolumeIntegrator = AutoVolumeProjection(),
        *,
        image_transform: AbstractImageTransform | None = None,
        normalizes_signal: bool = False,
        signal_region: Bool[NDArrayLike, "_ _"] | None = None,
        signal_centering: Literal["bg", "mean"] = "mean",
        translate_mode: Literal["fft", "atom", "none"] = "fft",
    ):
        """**Arguments:**

        - `volume`:
            The parametrization of the imaging volume
        - `pose`:
            The pose of the volume.
        - `image_config`:
            The configuration of the instrument, such as for the pixel size
            and the wavelength.
        - `volume_integrator`: The method for integrating the volume onto the plane.
        - `image_transform`:
            A [`cryojax.ndimage.AbstractImageTransform`][] applied to the
            image after simulation.
        - `normalizes_signal`:
            Whether or not to normalize the output of `image_model.simulate()`.
            If `True`, see `signal_centering` for options.
        - `signal_region`:
            A boolean array that is 1 where there is signal,
            and 0 otherwise used to normalize the image.
            Must have shape equal to `AbstractImageConfig.shape`.
        - `signal_centering`:
            How to calculate the offset for normalization when
            `normalizes_signal = True`. Options are
            - 'mean':
                Normalize the image to be mean 0
                within `signal_region`.
            - 'bg':
                Subtract mean value at the image edges.
                This makes the image fade to a background with values
                equal to zero. Requires that `image_config.padded_shape`
                is large enough so that the signal sufficiently decays.
            Ignored if `normalizes_signal = False`.
        - `translate_mode`:
            How to apply in-plane translation to the volume. Options are
            - 'fft':
                Apply phase shifts in the Fourier domain.
            - 'atom':
                Apply translation to atom positions before
                projection. For this method, the
                [`cryojax.simulator.AbstractVolumeParametrization`][]
                must be or return an [`cryojax.simulator.AbstractAtomVolume`][].
            - 'none':
                Do not apply the translation.
        """
        # Simulator components
        self.volume = volume
        self.pose = pose
        self.image_config = image_config
        self.volume_integrator = volume_integrator
        # Options
        self.image_transform = image_transform
        self.translate_mode = translate_mode
        self.normalizes_signal = normalizes_signal
        self.signal_centering = signal_centering
        if signal_region is None:
            self.signal_region = None
        else:
            self.signal_region = jnp.asarray(signal_region, dtype=bool)

    @override
    def raw_simulate(
        self,
        rng_key: PRNGKeyArray | None = None,
        *,
        outputs_real_space: bool = True,
    ) -> ImageArray | PaddedImageArray:
        # Get the representation of the volume
        if rng_key is None:
            volume_representation = self.volume.to_representation()
        else:
            this_key, rng_key = jr.split(rng_key)
            volume_representation = self.volume.to_representation(rng_key=this_key)
        # Rotate it to the lab frame
        volume_representation = volume_representation.rotate_to_pose(self.pose)
        # Translate if using atom translations
        if self.translate_mode == "atom":
            volume_representation = self._atom_translate(volume_representation)
        # Compute the projection image
        fourier_image = self.volume_integrator.integrate(
            volume_representation, self.image_config, outputs_real_space=False
        )
        # Now for the in-plane translation
        if self.translate_mode == "fft":
            fourier_image = self._phase_shift_translate(fourier_image)

        return (
            irfftn(fourier_image, s=self.image_config.padded_shape)
            if outputs_real_space
            else fourier_image
        )


class AbstractPhysicalImageModel(AbstractImageModel, strict=True):
    """An image formation model that simulates physical
    quantities. This uses the `AbstractScatteringTheory` class.
    """

    scattering_theory: eqx.AbstractVar[AbstractScatteringTheory]


class ContrastImageModel(AbstractPhysicalImageModel, strict=True):
    """An image formation model that returns the image contrast from a linear
    scattering theory.
    """

    volume: AbstractVolumeParametrization
    pose: AbstractPose
    image_config: AbstractImageConfig
    scattering_theory: AbstractScatteringTheory

    image_transform: AbstractImageTransform | None
    normalizes_signal: bool
    signal_region: Bool[Array, "_ _"] | None
    signal_centering: Literal["bg", "mean"]
    translate_mode: Literal["fft", "atom", "none"]

    def __init__(
        self,
        volume: AbstractVolumeParametrization,
        pose: AbstractPose,
        image_config: AbstractImageConfig,
        scattering_theory: AbstractScatteringTheory,
        *,
        image_transform: AbstractImageTransform | None = None,
        normalizes_signal: bool = False,
        signal_region: Bool[NDArrayLike, "_ _"] | None = None,
        signal_centering: Literal["bg", "mean"] = "mean",
        translate_mode: Literal["fft", "atom", "none"] = "fft",
    ):
        self.volume = volume
        self.pose = pose
        self.image_config = image_config
        self.scattering_theory = scattering_theory
        self.image_transform = image_transform
        self.translate_mode = translate_mode
        self.normalizes_signal = normalizes_signal
        self.signal_centering = signal_centering
        if signal_region is None:
            self.signal_region = None
        else:
            self.signal_region = jnp.asarray(signal_region, dtype=bool)

    @override
    def raw_simulate(
        self,
        rng_key: PRNGKeyArray | None = None,
        *,
        outputs_real_space: bool = True,
    ) -> PaddedFourierImageArray:
        # Get the volume representation. Its data should be a scattering potential
        # to simulate in physical units
        if rng_key is None:
            volume_representation = self.volume.to_representation()
        else:
            this_key, rng_key = jr.split(rng_key)
            volume_representation = self.volume.to_representation(rng_key=this_key)
        # Rotate it to the lab frame
        volume_representation = volume_representation.rotate_to_pose(self.pose)
        # Translate if using atom translations
        if self.translate_mode == "atom":
            volume_representation = self._atom_translate(volume_representation)
        # Compute the contrast
        contrast_spectrum = self.scattering_theory.compute_contrast_spectrum(
            volume_representation,
            self.image_config,
            rng_key,
            defocus_offset=self.pose.offset_z_in_angstroms,
        )
        # Apply the translation
        if self.translate_mode == "fft":
            contrast_spectrum = self._phase_shift_translate(contrast_spectrum)

        return (
            irfftn(contrast_spectrum, s=self.image_config.padded_shape)
            if outputs_real_space
            else contrast_spectrum
        )


class IntensityImageModel(AbstractPhysicalImageModel, strict=True):
    """An image formation model that returns an intensity distribution---or in other
    words a squared wavefunction.
    """

    volume: AbstractVolumeParametrization
    pose: AbstractPose
    image_config: AbstractImageConfig
    scattering_theory: AbstractScatteringTheory

    image_transform: AbstractImageTransform | None
    normalizes_signal: bool
    signal_region: Bool[Array, "_ _"] | None
    signal_centering: Literal["bg", "mean"]
    translate_mode: Literal["fft", "atom", "none"]

    def __init__(
        self,
        volume: AbstractVolumeParametrization,
        pose: AbstractPose,
        image_config: AbstractImageConfig,
        scattering_theory: AbstractScatteringTheory,
        *,
        image_transform: AbstractImageTransform | None = None,
        normalizes_signal: bool = False,
        signal_region: Bool[NDArrayLike, "_ _"] | None = None,
        signal_centering: Literal["bg", "mean"] = "mean",
        translate_mode: Literal["fft", "atom", "none"] = "fft",
    ):
        self.volume = volume
        self.pose = pose
        self.image_config = image_config
        self.scattering_theory = scattering_theory
        self.image_transform = image_transform
        self.translate_mode = translate_mode
        self.normalizes_signal = normalizes_signal
        self.signal_centering = signal_centering
        if signal_region is None:
            self.signal_region = None
        else:
            self.signal_region = jnp.asarray(signal_region, dtype=bool)

    @override
    def raw_simulate(
        self,
        rng_key: PRNGKeyArray | None = None,
        *,
        outputs_real_space: bool = True,
    ) -> PaddedFourierImageArray:
        # Get the volume representation. Its data should be a scattering potential
        # to simulate in physical units
        if rng_key is None:
            volume_representation = self.volume.to_representation()
        else:
            this_key, rng_key = jr.split(rng_key)
            volume_representation = self.volume.to_representation(rng_key=this_key)
        # Rotate it to the lab frame
        volume_representation = volume_representation.rotate_to_pose(self.pose)
        # Translate if using atom translations
        if self.translate_mode == "atom":
            volume_representation = self._atom_translate(volume_representation)
        # Compute the intensity spectrum
        intensity_spectrum = self.scattering_theory.compute_intensity_spectrum(
            volume_representation,
            self.image_config,
            rng_key,
            defocus_offset=self.pose.offset_z_in_angstroms,
        )
        if self.translate_mode == "fft":
            intensity_spectrum = self._phase_shift_translate(intensity_spectrum)

        return (
            irfftn(intensity_spectrum, s=self.image_config.padded_shape)
            if outputs_real_space
            else intensity_spectrum
        )


class ElectronCountsImageModel(AbstractPhysicalImageModel, strict=True):
    """An image formation model that returns electron counts, given a
    model for the detector.
    """

    volume: AbstractVolumeParametrization
    pose: AbstractPose
    image_config: DoseImageConfig
    scattering_theory: AbstractScatteringTheory
    detector: AbstractDetector

    image_transform: AbstractImageTransform | None
    normalizes_signal: bool
    signal_region: Bool[Array, "_ _"] | None
    signal_centering: Literal["bg", "mean"]
    translate_mode: Literal["fft", "atom", "none"]

    def __init__(
        self,
        volume: AbstractVolumeParametrization,
        pose: AbstractPose,
        image_config: DoseImageConfig,
        scattering_theory: AbstractScatteringTheory,
        detector: AbstractDetector,
        *,
        image_transform: AbstractImageTransform | None = None,
        normalizes_signal: bool = False,
        signal_region: Bool[NDArrayLike, "_ _"] | None = None,
        signal_centering: Literal["bg", "mean"] = "mean",
        translate_mode: Literal["fft", "atom", "none"] = "fft",
    ):
        self.volume = volume
        self.pose = pose
        self.image_config = image_config
        self.scattering_theory = scattering_theory
        self.detector = detector
        self.image_transform = image_transform
        self.translate_mode = translate_mode
        self.normalizes_signal = normalizes_signal
        self.signal_centering = signal_centering
        if signal_region is None:
            self.signal_region = None
        else:
            self.signal_region = jnp.asarray(signal_region, dtype=bool)

    @override
    def raw_simulate(
        self,
        rng_key: PRNGKeyArray | None = None,
        *,
        outputs_real_space: bool = True,
    ) -> PaddedFourierImageArray:
        if rng_key is None:
            # Get the volume representation. Its data should be a scattering potential
            # to simulate in physical units
            volume_representation = self.volume.to_representation()
            # Rotate it to the lab frame
            volume_representation = volume_representation.rotate_to_pose(self.pose)
            # Translate if using atom translations
            if self.translate_mode == "atom":
                volume_representation = self._atom_translate(volume_representation)
            # Compute the intensity
            fourier_intensity = self.scattering_theory.compute_intensity_spectrum(
                volume_representation,
                self.image_config,
                defocus_offset=self.pose.offset_z_in_angstroms,
            )
            if self.translate_mode == "fft":
                fourier_intensity = self._phase_shift_translate(fourier_intensity)
            # ... now measure the expected electron events at the detector
            fourier_expected_electron_events = (
                self.detector.compute_expected_electron_events(
                    fourier_intensity, self.image_config
                )
            )

            return (
                irfftn(fourier_expected_electron_events, s=self.image_config.padded_shape)
                if outputs_real_space
                else fourier_expected_electron_events
            )
        else:
            keys = jr.split(rng_key, 3)
            # Get the volume representation. Its data should be a scattering potential
            # to simulate in physical units
            volume_representation = self.volume.to_representation(keys[0])
            # Rotate it to the lab frame
            volume_representation = volume_representation.rotate_to_pose(self.pose)
            # Translate if using atom translations
            if self.translate_mode == "atom":
                volume_representation = self._atom_translate(volume_representation)
            # Compute the squared wavefunction
            fourier_intensity = self.scattering_theory.compute_intensity_spectrum(
                volume_representation,
                self.image_config,
                keys[1],
                defocus_offset=self.pose.offset_z_in_angstroms,
            )
            if self.translate_mode == "fft":
                fourier_intensity = self._phase_shift_translate(fourier_intensity)
            # ... now measure the detector readout
            fourier_detector_readout = self.detector.compute_detector_readout(
                keys[2],
                fourier_intensity,
                self.image_config,
            )

            return (
                irfftn(fourier_detector_readout, s=self.image_config.padded_shape)
                if outputs_real_space
                else fourier_detector_readout
            )


_init_doc = """**Arguments:**

- `volume`:
    The parametrization of the imaging volume.
- `pose`:
    The pose of the volume.
- `image_config`:
    The configuration of the instrument, such as for the pixel size
    and the wavelength.
- `scattering_theory`:
    The scattering theory.
- `image_transform`:
    A [`cryojax.ndimage.AbstractImageTransform`][] applied to the
    image after simulation.
- `normalizes_signal`:
    Whether or not to normalize the output of `image_model.simulate()`.
    If `True`, see `signal_centering` for options.
- `signal_region`:
    A boolean array that is 1 where there is signal,
    and 0 otherwise used to normalize the image.
    Must have shape equal to `AbstractImageConfig.shape`.
- `signal_centering`:
    How to calculate the offset for normalization when
    `normalizes_signal = True`. Options are
    - 'mean':
        Normalize the image to be mean 0
        within `signal_region`.
    - 'bg':
        Subtract mean value at the image edges.
        This makes the image fade to a background with values
        equal to zero. Requires that `image_config.padded_shape`
        is large enough so that the signal sufficiently decays.
    Ignored if `normalizes_signal = False`.
- `translate_mode`:
    How to apply in-plane translation to the volume. Options are
    - 'fft':
        Apply phase shifts in the Fourier domain.
    - 'atom':
        Apply translation to atom positions before
        projection. For this method, the
        [`cryojax.simulator.AbstractVolumeParametrization`][]
        must be or return an [`cryojax.simulator.AbstractAtomVolume`][].
    - 'none':
        Do not apply the translation.
"""

ContrastImageModel.__init__.__doc__ = _init_doc
IntensityImageModel.__init__.__doc__ = _init_doc
ElectronCountsImageModel.__init__.__doc__ = _init_doc



import cryojax.simulator as cs
import jax
import jax.numpy as jnp
import numpy as np
from cryojax.ndimage import irfftn, rfftn


def test_constant_wavefunction_gives_constant_expected_events():
    # Pick a large integrated electron flux to test
    # Create DoseImageConfig, picking a large electron flux to test
    image_config = cs.DoseImageConfig(
        (25, 25),
        1.0,
        voltage_in_kilovolts=300.0,
        electron_dose=10000.0,
    )
    electrons_per_pixel = image_config.electron_dose * image_config.pixel_size**2
    # Create squared wavefunction of just vacuum, i.e. 1 everywhere
    vacuum_squared_wavefunction = jnp.ones(image_config.shape, dtype=float)
    fourier_vacuum_squared_wavefunction = rfftn(vacuum_squared_wavefunction)
    # Create detector models
    dqe = cs.NullDQE()
    poisson_detector = cs.PoissonDetector(dqe)
    # Compute expected events
    fourier_expected_electron_events = poisson_detector.compute_expected_electron_events(
        fourier_vacuum_squared_wavefunction,
        image_config,
    )
    # Make sure it is equal to the electron per pixel
    np.testing.assert_allclose(
        irfftn(fourier_expected_electron_events, s=image_config.padded_shape),
        jnp.full(image_config.padded_shape, electrons_per_pixel),
        # rtol=1e-2,
    )


def test_gaussian_limit():
    # Pick a large integrated electron flux to test
    # Create DoseImageConfig, picking a large electron flux to test
    image_config = cs.DoseImageConfig(
        (25, 25),
        1.0,
        voltage_in_kilovolts=300.0,
        electron_dose=10000.0,
    )
    n_pixels = np.prod(image_config.padded_shape)
    electrons_per_pixel = image_config.electron_dose * image_config.pixel_size**2
    # Create squared wavefunction of just vacuum, i.e. 1 everywhere
    vacuum_squared_wavefunction = jnp.ones(image_config.shape, dtype=float)
    fourier_vacuum_squared_wavefunction = rfftn(vacuum_squared_wavefunction)
    # Create detector models
    key = jax.random.key(1234)
    dqe = cs.NullDQE()
    gaussian_detector = cs.GaussianDetector(dqe)
    poisson_detector = cs.PoissonDetector(dqe)
    # Compute detector readout
    fourier_gaussian_detector_readout = gaussian_detector.compute_detector_readout(
        key,
        fourier_vacuum_squared_wavefunction,
        image_config,
    )
    fourier_poisson_detector_readout = poisson_detector.compute_detector_readout(
        key,
        fourier_vacuum_squared_wavefunction,
        image_config,
    )
    # Compare to see if the autocorrelation has converged
    np.testing.assert_allclose(
        irfftn(
            jnp.abs(fourier_gaussian_detector_readout) ** 2
            / (n_pixels * electrons_per_pixel**2),
            s=image_config.padded_shape,
        ),
        irfftn(
            jnp.abs(fourier_poisson_detector_readout) ** 2
            / (n_pixels * electrons_per_pixel**2),
            s=image_config.padded_shape,
        ),
        rtol=1e-2,
    )



"""
Routines to compute radial averages of images.
"""

import jax.numpy as jnp
from jaxtyping import Array, Float, Inexact


def compute_binned_radial_average(
    image: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"],
    radial_coordinate_grid: (
        Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]
    ),
    bins: Float[Array, " n_bins"],
) -> Inexact[Array, " n_bins"]:
    """Average vectors $\\mathbf{r}$ of constant radius $|\\mathbf{r}|$ into
    discrete bins.

    **Arguments:**

    - `image`:
        Two-dimensional image or three-dimensional volume.
    - `radial_coordinate_grid`:
        Radial coordinate system of image or volume.
    - `bins`:
        Radial bins for averaging.

    **Returns:**

    The binned radial averaged of `image` in bins `bins`.
    """
    # Discretize the radial grid
    digitized_radial_grid = jnp.digitize(radial_coordinate_grid, bins, right=True)
    # Compute the radial profile as the average value of the image in each bin
    binned_radial_average = jnp.bincount(
        digitized_radial_grid.ravel(),
        weights=image.ravel(),
        length=bins.size,
    ) / jnp.bincount(digitized_radial_grid.ravel(), length=bins.size)

    return binned_radial_average


def radial_average_to_grid(
    binned_radial_average: Inexact[Array, " n_bins"],
    bins: Float[Array, " n_bins"],
    radial_coordinate_grid: (
        Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]
    ),
    interpolation_mode: str = "linear",
) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
    """Interpolate a binned radially averaged profile onto a grid.

    **Arguments:**

    - `binned_radial_average`:
        The binned, radially averaged profile.
    - `bins`:
        Radial bins over which `binned_radial_average` is computed.
    - `radial_coordinate_grid`:
        Radial coordinate system of image or volume.
    - `interpolation_mode`:
        If `"linear"`, evaluate the grid using linear
        interpolation. If `"nearest"`, use nearest-neighbor
        interpolation.

    **Returns:**

    The `binned_radial_average` evaluated on the `radial_coordinate_grid`.
    """
    if interpolation_mode == "nearest":
        radial_average_on_grid = jnp.take(
            binned_radial_average,
            jnp.digitize(radial_coordinate_grid, bins, right=True),
            mode="clip",
        )
    elif interpolation_mode == "linear":
        radial_average_on_grid = jnp.interp(
            radial_coordinate_grid.ravel(),
            bins,
            binned_radial_average,
        ).reshape(radial_coordinate_grid.shape)
    else:
        raise ValueError(
            f"`interpolation_mode` = {interpolation_mode} not supported. Supported "
            "interpolation modes are 'nearest' or 'linear'."
        )
    return radial_average_on_grid



from ._fourier_operator import (
    AbstractFourierOperator as AbstractFourierOperator,
    CustomFourierOperator as CustomFourierOperator,
    FourierConstant as FourierConstant,
    FourierDC as FourierDC,
    FourierExp2D as FourierExp2D,
    FourierGaussian as FourierGaussian,
    FourierPhaseShifts as FourierPhaseShifts,
    FourierSinc as FourierSinc,
    PeakedFourierGaussian as PeakedFourierGaussian,
)
from ._real_operator import (
    AbstractRealOperator as AbstractRealOperator,
    RealConstant as RealConstant,
    RealGaussian as RealGaussian,
)



from .multislice_integrator import (
    AbstractMultisliceIntegrator as AbstractMultisliceIntegrator,
    FFTMultisliceIntegrator as FFTMultisliceIntegrator,
)
from .multislice_scattering_theory import (
    MultisliceScatteringTheory as MultisliceScatteringTheory,
)



import numpy as np
from jaxtyping import Float


def center_atom_positions(
    atom_positions: Float[np.ndarray, "... n_atoms 3"],
    atom_masses: Float[np.ndarray, "... n_atoms"] | None = None,
    atom_axis: int = 0,
) -> Float[np.ndarray, "... n_atoms 3"]:
    """Center positions of a numpy array of atoms.

    **Arguments:**

    - `atom_positions`:
        The array of atom positions.
    - `atom_masses`:
        Optionally, provide the masses of the atoms
        to center by mass.
    - `atom_axis`:
        The axis representing the atom index. Leading axes
        are assumed to be batched axes. `atom_positions`
        and `atom_masses` are assumed to have the same
        batched axes.

    **Returns:**

    The centered `atom_positions`. This is centered by
    geometry if `atom_masses = None` and by mass otherwise.
    """
    if atom_masses is None:
        n_atoms = atom_positions.shape[atom_axis]
        center_position = np.sum(atom_positions, axis=atom_axis) / n_atoms
    else:
        center_position = (
            np.sum(atom_positions * atom_masses[..., None], axis=atom_axis)
            / atom_masses.sum(axis=atom_axis)[..., None]
        )
    return atom_positions - np.expand_dims(center_position, axis=atom_axis)



import jax.numpy as jnp
from jaxtyping import Array, Float


def convert_quaternion_to_euler_angles(
    wxyz: Float[Array, "4"], convention: str = "zyz", extrinsic: bool = True
) -> Float[Array, "3"]:
    """Convert a quaternion to a sequence of euler angles about an extrinsic
    coordinate system.

    Adapted from https://github.com/chrisflesher/jax-scipy-spatial/.
    """
    if len(convention) != 3 or not all([axis in ["x", "y", "z"] for axis in convention]):
        raise ValueError(
            f"`convention` should be a string of three characters, in "
            f"of which is 'x', 'y', or 'z'. Instead, got '{convention}'"
        )
    if convention[0] == convention[1] or convention[1] == convention[2]:
        raise ValueError(
            f"`convention` cannot have axes repeating in a row. For example, "
            f"'xxy' or 'zzz' are not allowed. Got '{convention}'."
        )
    if convention[0] != convention[2]:
        raise ValueError(
            f"`convention` must have a repeated first and last axis. For example, "
            f"'zyz' or 'xyx' are allowed. Got '{convention}'."
        )
    xyz_axis_to_array_axis = {"x": 0, "y": 1, "z": 2}
    axes = [xyz_axis_to_array_axis[axis] for axis in convention]
    angle_first, angle_third = (0, 2) if extrinsic else (2, 0)
    axes = axes if extrinsic else axes[::-1]
    xyzw = jnp.roll(wxyz, shift=-1)
    i, j, k = axes
    eps = 1e-7
    k = 3 - i - j
    sign = jnp.array((i - j) * (j - k) * (k - i) // 2, dtype=xyzw.dtype)
    a, b, c, d = (xyzw[3], xyzw[i], xyzw[j], xyzw[k] * sign)
    angles = jnp.empty(3, dtype=xyzw.dtype)
    angles = angles.at[1].set(2 * jnp.arctan2(jnp.hypot(c, d), jnp.hypot(a, b)))
    case = jnp.where(jnp.abs(angles[1] - jnp.pi) <= eps, 2, 0)
    case = jnp.where(jnp.abs(angles[1]) <= eps, 1, case)
    half_sum = jnp.arctan2(b, a)
    half_diff = jnp.arctan2(d, c)
    angles = angles.at[0].set(
        jnp.where(case == 1, 2 * half_sum, 2 * half_diff * -1)
    )  # any degenerate case
    angles = angles.at[angle_first].set(
        jnp.where(case == 0, half_sum - half_diff, angles[angle_first])
    )
    angles = angles.at[angle_third].set(
        jnp.where(case == 0, half_sum + half_diff, angles[angle_third])
    )
    angles = (angles + jnp.pi) % (2 * jnp.pi) - jnp.pi
    return jnp.rad2deg(angles)



"""The contents here are not public API, but are used internally throughout
cryoJAX.
"""

from collections.abc import Callable

import jax
import jax.numpy as jnp
from jax.debug import callback
from jaxtyping import Array

from .jax_util import maybe_error_if


#
# Helpers for performing internal error checks
#
_make_msg = (
    lambda _s: "While inspecting runtime errors with `CRYOJAX_ENABLE_CHECKS=true`, "
    + _s
    + (
        " Inspect the traceback to determine where this occurred, or "
        "set `EQX_ON_ERROR=breakpoint` and use a debugger."
    )
)


def error_if_negative(x: Array) -> Array:
    return maybe_error_if(
        x,
        lambda _x: _x < 0,
        _make_msg("a non-negative quantity was found to be negative."),
    )


def error_if_not_positive(x: Array) -> Array:
    return maybe_error_if(
        x,
        lambda _x: _x <= 0,
        _make_msg("a positive quantity was found to be negative or zero."),
    )


def error_if_zero(x: Array) -> Array:
    return maybe_error_if(
        x,
        lambda _x: jnp.isclose(_x, 0.0),
        _make_msg("a non-zero quantity was found to be zero."),
    )


def error_if_not_fractional(x: Array) -> Array:
    return maybe_error_if(
        x,
        lambda _x: ~jnp.logical_and(_x >= 0.0, _x <= 1.0),
        _make_msg("a fractional quantity was found to not be between 0 and 1."),
    )


#
# Miscellaneous
#
def fori_loop_tqdm_decorator(
    n_iterations: int,
    print_every: int | None = None,
    **kwargs,
) -> Callable:
    """Add a tqdm progress bar to `body_fun` used in `jax.lax.fori_loop`.
    This function is closely based on the implementation in
    [`jax_tqdm`](https://github.com/jeremiecoullon/jax-tqdm).
    """

    _update_progress_bar, close_tqdm = _build_tqdm(n_iterations, print_every, **kwargs)

    def _fori_loop_tqdm_decorator(func):
        def wrapper_progress_bar(i, val):
            _update_progress_bar(i)
            result = func(i, val)
            return close_tqdm(result, i)

        return wrapper_progress_bar

    return _fori_loop_tqdm_decorator


def _build_tqdm(
    n: int, print_rate: int | None = None, **kwargs
) -> tuple[Callable, Callable]:
    from tqdm.auto import tqdm as pbar

    desc = kwargs.pop("desc", f"Running for {n:,} iterations")
    message = kwargs.pop("message", desc)
    for kwarg in ("total", "mininterval", "maxinterval", "miniters"):
        kwargs.pop(kwarg, None)

    tqdm_bars = {}

    if print_rate is None:
        if n > 20:
            print_rate = int(n / 20)
        else:
            print_rate = 1
    else:
        if print_rate < 1:
            raise ValueError(f"Print rate should be > 0 got {print_rate}")
        elif print_rate > n:
            raise ValueError(
                "Print rate should be less than the "
                f"number of steps {n}, got {print_rate}"
            )

    remainder = n % print_rate

    def _define_tqdm(arg, transform):
        tqdm_bars[0] = pbar(range(n), **kwargs)
        tqdm_bars[0].set_description(message, refresh=False)

    def _update_tqdm(arg, transform):
        tqdm_bars[0].update(int(arg))

    def _update_progress_bar(iter_num):
        "Updates tqdm from a JAX scan or loop"
        _ = jax.lax.cond(
            iter_num == 0,
            lambda _: callback(_define_tqdm, None, None, ordered=True),
            lambda _: None,
            operand=None,
        )

        _ = jax.lax.cond(
            # update tqdm every multiple of `print_rate` except at the end
            (iter_num % print_rate == 0) & (iter_num != n - remainder),
            lambda _: callback(_update_tqdm, print_rate, None, ordered=True),
            lambda _: None,
            operand=None,
        )

        _ = jax.lax.cond(
            # update tqdm by `remainder`
            iter_num == n - remainder,
            lambda _: callback(_update_tqdm, remainder, None, ordered=True),
            lambda _: None,
            operand=None,
        )

    def _close_tqdm(arg, transform):
        tqdm_bars[0].close()

    def close_tqdm(result, iter_num):
        _ = jax.lax.cond(
            iter_num == n - 1,
            lambda _: callback(_close_tqdm, None, None, ordered=True),
            lambda _: None,
            operand=None,
        )
        return result

    return _update_progress_bar, close_tqdm



"""
Fourier voxel-based representations of a volume.
"""

from typing import ClassVar, Literal, cast
from typing_extensions import Self, override

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Complex, Float

from ...jax_util import NDArrayLike
from ...ndimage import (
    SincCorrectionMask,
    compute_spline_coefficients,
    convert_fftn_to_rfftn,
    fftn,
    ifftn,
    irfftn,
    make_coordinate_grid,
    make_frequency_slice,
    map_coordinates,
    map_coordinates_spline,
    pad_to_shape,
    resize_with_crop_or_pad,
    rfftn,
)
from .._image_config import AbstractImageConfig
from .._pose import AbstractPose
from .base_volume import (
    AbstractVolumeIntegrator,
    AbstractVoxelVolume,
    EwaldSphereArray,
    ProjectionArray,
)


class AbstractFourierVoxelVolume(AbstractVoxelVolume, strict=True):
    """Abstract interface for a voxel-based volume."""

    frequency_slice_in_pixels: eqx.AbstractVar[Float[Array, "1 dim dim 3"]]

    @override
    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Return a new volume with a rotated `frequency_slice_in_pixels`."""
        return eqx.tree_at(
            lambda d: d.frequency_slice_in_pixels,
            self,
            pose.rotate_coordinates(self.frequency_slice_in_pixels, inverse=inverse),
        )


class FourierVoxelGridVolume(AbstractFourierVoxelVolume, strict=True):
    """A 3D voxel grid in fourier-space."""

    fourier_voxel_grid: Complex[Array, "dim dim dim"]
    frequency_slice_in_pixels: Float[Array, "1 dim dim 3"]

    rotation_convention: ClassVar[Literal["frame"]] = "frame"

    def __init__(
        self,
        fourier_voxel_grid: Complex[NDArrayLike, "dim dim dim"],
        frequency_slice_in_pixels: Float[NDArrayLike, "1 dim dim 3"],
    ):
        """**Arguments:**

        - `fourier_voxel_grid`:
            The cubic voxel grid in fourier space.
        - `frequency_slice_in_pixels`:
            The frequency slice coordinate system.
        """
        self.fourier_voxel_grid = jnp.asarray(fourier_voxel_grid, dtype=complex)
        self.frequency_slice_in_pixels = jnp.asarray(
            frequency_slice_in_pixels, dtype=float
        )

    @property
    def shape(self) -> tuple[int, int, int]:
        """The shape of the `fourier_voxel_grid`."""
        return cast(tuple[int, int, int], self.fourier_voxel_grid.shape)

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[NDArrayLike, "dim dim dim"],
        *,
        sinc_correction: bool = False,
        pad_scale: float = 1.0,
        pad_mode: str = "constant",
    ) -> Self:
        """Load from a real-valued 3D voxel grid.

        **Arguments:**

        - `real_voxel_grid`:
            A voxel grid in real space.
        - `sinc_correction`:
            If `True`, apply a [`cryojax.ndimage.SincCorrectionMask`][]
            to correct for errors incurred from linear interpolation in
            the Fourier domain via [`cryojax.simulator.FourierSliceExtraction`][].
        - `pad_scale`:
            Scale factor at which to pad `real_voxel_grid` before fourier
            transform. Must be a value greater than `1.0`.
        - `pad_mode`:
            Padding method. See `jax.numpy.pad` for documentation.
        """
        # Cast to jax array
        real_voxel_grid = jnp.asarray(real_voxel_grid, dtype=float)
        # Pad template
        if pad_scale < 1.0:
            raise ValueError("`pad_scale` must be greater than 1.0")
        # ... always pad to even size to avoid interpolation issues in
        # fourier slice extraction.
        padded_shape = cast(
            tuple[int, int, int],
            tuple([int(s * pad_scale) for s in real_voxel_grid.shape]),
        )
        padded_real_voxel_grid = pad_to_shape(
            real_voxel_grid, padded_shape, mode=pad_mode
        )
        if sinc_correction:
            coordinate_grid = make_coordinate_grid(padded_shape)
            correction_mask = SincCorrectionMask(coordinate_grid)
            padded_real_voxel_grid = correction_mask(padded_real_voxel_grid)
        # Load grid and coordinates. For now, do not store the
        # fourier grid only on the half space. Fourier slice extraction
        # does not currently work if rfftn is used.
        fourier_voxel_grid_with_zero_in_corner = fftn(padded_real_voxel_grid)
        # ... store the grid with the zero frequency component in the center
        fourier_voxel_grid = jnp.fft.fftshift(fourier_voxel_grid_with_zero_in_corner)
        # ... create in-plane frequency slice on the half space
        frequency_slice = make_frequency_slice(
            cast(tuple[int, int], padded_real_voxel_grid.shape[:-1]),
            outputs_rfftfreqs=False,
        )

        return cls(fourier_voxel_grid, frequency_slice)


class FourierVoxelSplineVolume(AbstractFourierVoxelVolume, strict=True):
    """A 3D voxel grid in fourier-space, represented
    by spline coefficients.
    """

    spline_coefficients: Complex[Array, "coeff_dim coeff_dim coeff_dim"]
    frequency_slice_in_pixels: Float[Array, "1 dim dim 3"]

    rotation_convention: ClassVar[Literal["frame"]] = "frame"

    def __init__(
        self,
        spline_coefficients: Complex[NDArrayLike, "coeff_dim coeff_dim coeff_dim"],
        frequency_slice_in_pixels: Float[NDArrayLike, "1 dim dim 3"],
    ):
        """**Arguments:**

        - `spline_coefficients`:
            The spline coefficents computed from the cubic voxel grid
            in fourier space. See `cryojax.ndimage.compute_spline_coefficients`.
        - `frequency_slice_in_pixels`:
            Frequency slice coordinate system.
            See `cryojax.coordinates.make_frequency_slice`.
        """
        self.spline_coefficients = jnp.asarray(spline_coefficients, dtype=complex)
        self.frequency_slice_in_pixels = jnp.asarray(
            frequency_slice_in_pixels, dtype=float
        )

    @property
    def shape(self) -> tuple[int, int, int]:
        """The shape of the original `fourier_voxel_grid` from which
        `coefficients` were computed.
        """
        return cast(
            tuple[int, int, int], tuple([s - 2 for s in self.spline_coefficients.shape])
        )

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[NDArrayLike, "dim dim dim"],
        *,
        pad_scale: float = 1.0,
        pad_mode: str = "constant",
    ) -> Self:
        """Load from a real-valued 3D voxel grid.

        **Arguments:**

        - `real_voxel_grid`: A voxel grid in real space.
        - `pad_scale`: Scale factor at which to pad `real_voxel_grid` before fourier
                     transform. Must be a value greater than `1.0`.
        - `pad_mode`: Padding method. See `jax.numpy.pad` for documentation.
        """
        # Cast to jax array
        real_voxel_grid = jnp.asarray(real_voxel_grid, dtype=float)
        # Pad template
        if pad_scale < 1.0:
            raise ValueError("`pad_scale` must be greater than 1.0")
        # ... always pad to even size to avoid interpolation issues in
        # fourier slice extraction.
        padded_shape = cast(
            tuple[int, int, int],
            tuple([int(s * pad_scale) for s in real_voxel_grid.shape]),
        )
        padded_real_voxel_grid = pad_to_shape(
            real_voxel_grid, padded_shape, mode=pad_mode
        )
        # Load grid and coordinates. For now, do not store the
        # fourier grid only on the half space. Fourier slice extraction
        # does not currently work if rfftn is used.
        fourier_voxel_grid_with_zero_in_corner = fftn(padded_real_voxel_grid)
        # ... store the grid with the zero frequency component in the center
        fourier_voxel_grid = jnp.fft.fftshift(fourier_voxel_grid_with_zero_in_corner)
        # ... compute spline coefficients
        spline_coefficients = compute_spline_coefficients(fourier_voxel_grid)
        # ... create in-plane frequency slice on the half space
        frequency_slice = make_frequency_slice(
            cast(tuple[int, int], padded_real_voxel_grid.shape[:-1]),
            outputs_rfftfreqs=False,
        )

        return cls(spline_coefficients, frequency_slice)


class FourierSliceExtraction(
    AbstractVolumeIntegrator[FourierVoxelGridVolume | FourierVoxelSplineVolume],
    strict=True,
):
    """Integrate points to the exit plane using the Fourier
    projection-slice theorem.

    This extracts slices using interpolation methods housed in
    `cryojax.ndimage.map_coordinates` and
    `cryojax.ndimage.map_coordinates_spline`.
    """

    outputs_integral: bool
    out_of_bounds_mode: str
    fill_value: complex

    outputs_ewald_sphere: ClassVar[bool] = False

    def __init__(
        self,
        *,
        outputs_integral: bool = True,
        out_of_bounds_mode: str = "fill",
        fill_value: complex = 0.0 + 0.0j,
    ):
        """**Arguments:**

        - `outputs_integral`:
            If `True`, return the fourier slice
            *multiplied by the voxel size*. Including the voxel size
            numerical approximates the projection integral and is
            necessary for simulating images in physical units.
        - `out_of_bounds_mode`:
            Specify how to handle out of bounds indexing. See
            `cryojax.ndimage.map_coordinates` for documentation.
        - `fill_value`:
            Value for filling out-of-bounds indices. Used only when
            `out_of_bounds_mode = "fill"`.
        """
        self.outputs_integral = outputs_integral
        self.out_of_bounds_mode = out_of_bounds_mode
        self.fill_value = fill_value

    @override
    def integrate(
        self,
        volume_representation: FourierVoxelGridVolume | FourierVoxelSplineVolume,
        image_config: AbstractImageConfig,
        outputs_real_space: bool = False,
    ) -> ProjectionArray:
        """Integrate the volume at the `AbstractImageConfig` settings
        of a voxel-based representation in fourier-space,
        using fourier slice extraction.

        **Arguments:**

        - `volume_representation`:
            The volume representation.
        - `image_config`:
            The image configuration.
        - `outputs_real_space`:
            If `True`, return the image in real space. Otherwise,
            return in Fourier.

        **Returns:**

        The volume projection in real or Fourier space at the
        `AbstractImageConfig.padded_shape` and the `image_config.pixel_size`.
        """
        frequency_slice = volume_representation.frequency_slice_in_pixels
        N = frequency_slice.shape[1]
        if volume_representation.shape != (N, N, N):
            raise AttributeError(
                "Only cubic boxes are supported for fourier slice extraction."
            )
        # Compute the fourier projection
        if isinstance(volume_representation, FourierVoxelSplineVolume):
            fourier_projection = _extract_slice_spline(
                volume_representation.spline_coefficients,
                frequency_slice,
                mode=self.out_of_bounds_mode,
                cval=self.fill_value,
            )
        elif isinstance(volume_representation, FourierVoxelGridVolume):
            fourier_projection = _extract_slice(
                volume_representation.fourier_voxel_grid,
                frequency_slice,
                interpolation_order=1,
                mode=self.out_of_bounds_mode,
                cval=self.fill_value,
            )
        else:
            raise ValueError(
                "Supported types for `volume_representation` are "
                "`FourierVoxelGridVolume` and FourierVoxelSplineVolume`."
            )

        # Resize the image to match the AbstractImageConfig.padded_shape
        if image_config.padded_shape != (N, N):
            fourier_projection = rfftn(
                resize_with_crop_or_pad(
                    irfftn(fourier_projection, s=(N, N)), image_config.padded_shape
                )
            )
        # Scale by voxel size to convert from projection to integral
        if self.outputs_integral:
            fourier_projection *= image_config.pixel_size
        return (
            irfftn(fourier_projection, s=image_config.padded_shape)
            if outputs_real_space
            else fourier_projection
        )


class EwaldSphereExtraction(
    AbstractVolumeIntegrator[FourierVoxelGridVolume | FourierVoxelSplineVolume],
    strict=True,
):
    """Integrate points to the exit plane by extracting a surface of
    the ewald sphere in fourier space.

    This extracts surfaces using interpolation methods housed in
    `cryojax.image.map_coordinates`
    and `cryojax.image.map_coordinates_spline`.
    """

    outputs_integral: bool
    out_of_bounds_mode: str
    fill_value: complex

    outputs_ewald_sphere: ClassVar[bool] = True

    def __init__(
        self,
        *,
        outputs_integral: bool = True,
        out_of_bounds_mode: str = "fill",
        fill_value: complex = 0.0 + 0.0j,
    ):
        """**Arguments:**

        - `outputs_integral`:
            If `True`, return the ewald sphere surface
            *multiplied by the voxel size*. Including the voxel size
            numerical approximates the projection integral and is
            necessary for simulating images in physical units.
        - `out_of_bounds_mode`:
            Specify how to handle out of bounds indexing. See
            `cryojax.image.map_coordinates` for documentation.
        - `fill_value`:
            Value for filling out-of-bounds indices. Used only when
            `out_of_bounds_mode = "fill"`.
        """
        self.outputs_integral = outputs_integral
        self.out_of_bounds_mode = out_of_bounds_mode
        self.fill_value = fill_value

    @override
    def integrate(
        self,
        volume_representation: FourierVoxelGridVolume | FourierVoxelSplineVolume,
        image_config: AbstractImageConfig,
        outputs_real_space: bool = False,
    ) -> EwaldSphereArray:
        """Integrate the volume at the `AbstractImageConfig` settings
        of a voxel-based representation in fourier-space, using fourier
        slice extraction.

        **Arguments:**

        - `volume_representation`:
            The volume representation.
        - `image_config`:
            The image configuration.
        - `outputs_real_space`:
            If `True`, return the Ewald sphere surface in
            real space. Otherwise, return in Fourier.

        **Returns:**

        The Ewald sphere surface in real or Fourier space at the
        `AbstractImageConfig.padded_shape` and the `image_config.pixel_size`.
        """
        frequency_slice = volume_representation.frequency_slice_in_pixels
        N = frequency_slice.shape[1]
        if volume_representation.shape != (N, N, N):
            raise AttributeError(
                "Only cubic boxes are supported for fourier slice extraction."
            )
        # Compute the fourier projection
        if isinstance(volume_representation, FourierVoxelSplineVolume):
            ewald_sphere_surface = _extract_ewald_sphere_spline(
                volume_representation.spline_coefficients,
                frequency_slice,
                image_config.pixel_size,
                image_config.wavelength_in_angstroms,
                mode=self.out_of_bounds_mode,
                cval=self.fill_value,
            )
        elif isinstance(volume_representation, FourierVoxelGridVolume):
            ewald_sphere_surface = _extract_ewald_sphere(
                volume_representation.fourier_voxel_grid,
                frequency_slice,
                image_config.pixel_size,
                image_config.wavelength_in_angstroms,
                interpolation_order=1,
                mode=self.out_of_bounds_mode,
                cval=self.fill_value,
            )
        else:
            raise ValueError(
                "Supported types for `volume_representation` are "
                "`FourierVoxelGridVolume` and `FourierVoxelSplineVolume`."
            )

        # Resize the image to match the AbstractImageConfig.padded_shape
        if image_config.padded_shape != (N, N):
            ewald_sphere_surface = fftn(
                resize_with_crop_or_pad(
                    ifftn(ewald_sphere_surface, s=(N, N)), image_config.padded_shape
                )
            )
        # Scale by voxel size to convert from projection to integral
        if self.outputs_integral:
            ewald_sphere_surface *= image_config.pixel_size
        return (
            irfftn(ewald_sphere_surface, s=image_config.padded_shape)
            if outputs_real_space
            else ewald_sphere_surface
        )


def _extract_slice(
    fourier_voxel_grid,
    frequency_slice,
    interpolation_order,
    **kwargs,
) -> Complex[Array, "dim dim//2+1"]:
    return convert_fftn_to_rfftn(
        _extract_surface_from_voxel_grid(
            fourier_voxel_grid,
            frequency_slice,
            is_spline_coefficients=False,
            interpolation_order=interpolation_order,
            **kwargs,
        ),
        mode="real",
    )


def _extract_slice_spline(
    spline_coefficients, frequency_slice, **kwargs
) -> Complex[Array, "dim dim//2+1"]:
    return convert_fftn_to_rfftn(
        _extract_surface_from_voxel_grid(
            spline_coefficients, frequency_slice, is_spline_coefficients=True, **kwargs
        ),
        mode="real",
    )


def _extract_ewald_sphere(
    fourier_voxel_grid,
    frequency_slice,
    voxel_size,
    wavelength,
    interpolation_order,
    **kwargs,
) -> Complex[Array, "dim dim"]:
    ewald_sphere_frequencies = _get_ewald_sphere_surface_from_slice(
        frequency_slice, voxel_size, wavelength
    )
    return _extract_surface_from_voxel_grid(
        fourier_voxel_grid,
        ewald_sphere_frequencies,
        is_spline_coefficients=False,
        interpolation_order=interpolation_order,
        **kwargs,
    )


def _extract_ewald_sphere_spline(
    spline_coefficients, frequency_slice, voxel_size, wavelength, **kwargs
) -> Complex[Array, "dim dim"]:
    ewald_sphere_frequencies = _get_ewald_sphere_surface_from_slice(
        frequency_slice, voxel_size, wavelength
    )
    return _extract_surface_from_voxel_grid(
        spline_coefficients,
        ewald_sphere_frequencies,
        is_spline_coefficients=True,
        **kwargs,
    )


def _get_ewald_sphere_surface_from_slice(
    frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
    voxel_size: Float[Array, ""],
    wavelength: Float[Array, ""],
) -> Float[Array, "1 dim dim 3"]:
    frequency_slice_with_zero_in_corner = jnp.fft.ifftshift(
        frequency_slice_in_pixels, axes=(0, 1, 2)
    )
    # Get zhat unit vector of the frequency slice
    xhat, yhat = (
        frequency_slice_with_zero_in_corner[0, 0, 1, :],
        frequency_slice_with_zero_in_corner[0, 1, 0, :],
    )
    xhat, yhat = xhat / jnp.linalg.norm(xhat), yhat / jnp.linalg.norm(yhat)
    zhat = jnp.cross(xhat, yhat)
    # Compute the ewald sphere surface, assuming the frequency slice is
    # in a rotated frame
    q_at_slice = frequency_slice_in_pixels
    q_squared = jnp.sum(q_at_slice**2, axis=-1)
    q_at_surface = (
        q_at_slice
        + (wavelength / voxel_size)
        * (q_squared[..., None] * zhat[None, None, None, :])
        / 2
    )
    return q_at_surface


def _extract_surface_from_voxel_grid(
    voxel_grid,
    frequency_coordinates,
    is_spline_coefficients=False,
    interpolation_order=1,
    **kwargs,
):
    # Convert to logical coordinates
    N = frequency_coordinates.shape[1]
    logical_frequency_coordinates = (frequency_coordinates * N) + N // 2
    # Convert arguments to map_coordinates convention and compute
    k_x, k_y, k_z = jnp.transpose(logical_frequency_coordinates, axes=[3, 0, 1, 2])
    if is_spline_coefficients:
        spline_coefficients = voxel_grid
        surface = map_coordinates_spline(spline_coefficients, (k_z, k_y, k_x), **kwargs)[
            0, :, :
        ]
    else:
        fourier_voxel_grid = voxel_grid
        surface = map_coordinates(
            fourier_voxel_grid, (k_z, k_y, k_x), interpolation_order, **kwargs
        )[0, :, :]
    # Shift zero frequency component to corner
    surface = jnp.fft.ifftshift(surface)

    return surface



"""
Routines to compute FFTs, in cryojax conventions.
"""

import jax.numpy as jnp
from jaxtyping import Array, Complex, Float, Inexact

from ..jax_util import NDArrayLike


def ifftn(
    ft: Inexact[NDArrayLike, "..."],
    s: tuple[int, ...] | None = None,
    axes: tuple[int, ...] | None = None,
) -> Complex[Array, "..."]:
    """The equivalent of `jax.numpy.fft.ifftn` in `cryojax` conventions.

    Arguments
    ---------
    ft :
        Fourier transform array. Assumes that the zero
        frequency component is in the corner.

    Returns
    -------
    ift :
        Inverse fourier transform.
    """
    ift = jnp.fft.fftshift(jnp.fft.ifftn(ft, s=s, axes=axes), axes=axes)

    return ift


def fftn(
    ift: Inexact[NDArrayLike, "..."],
    s: tuple[int, ...] | None = None,
    axes: tuple[int, ...] | None = None,
) -> Complex[Array, "..."]:
    """The equivalent of `jax.numpy.fft.fftn` in `cryojax` conventions.

    Arguments
    ---------
    ift :
        Array in real space. Assumes that the zero
        frequency component is in the center.

    Returns
    -------
    ft :
        Fourier transform of array.
    """
    ft = jnp.fft.fftn(jnp.fft.ifftshift(ift, axes=axes), s=s, axes=axes)

    return ft


def irfftn(
    ft: Inexact[NDArrayLike, "..."],
    s: tuple[int, ...] | None = None,
    axes: tuple[int, ...] | None = None,
) -> Float[Array, "..."]:
    """The equivalent of `jax.numpy.fft.irfftn` in `cryojax` conventions.

    Arguments
    ---------
    ft :
        Fourier transform array. Assumes that the zero
        frequency component is in the corner.

    Returns
    -------
    ift :
        Inverse fourier transform.
    """
    ift = jnp.fft.fftshift(jnp.fft.irfftn(ft, s=s, axes=axes), axes=axes)

    return ift


def rfftn(
    ift: Float[NDArrayLike, "..."],
    s: tuple[int, ...] | None = None,
    axes: tuple[int, ...] | None = None,
) -> Complex[Array, "..."]:
    """The equivalent of `jax.numpy.fft.rfftn` in `cryojax` conventions.

    Arguments
    ---------
    ift :
        Array in real space.

    Returns
    -------
    ft :
        Fourier transform of array.
    """
    ft = jnp.fft.rfftn(jnp.fft.ifftshift(ift, axes=axes), s=s, axes=axes)

    return ft



"""
Base class for a cryojax distribution.
"""

from abc import abstractmethod

import equinox as eqx
import jax
from jaxtyping import Array, Float, Inexact, PRNGKeyArray

from ...ndimage import AbstractFilter, AbstractMask
from .._detector import AbstractDetector
from .._solvent_2d import AbstractRandomSolvent2D


class AbstractNoiseModel(eqx.Module, strict=True):
    """An image formation model equipped with a noise model."""

    @abstractmethod
    def sample(
        self,
        rng_key: PRNGKeyArray,
        *,
        outputs_real_space: bool = True,
        mask: AbstractMask | None = None,
        filter: AbstractFilter | None = None,
    ) -> Inexact[Array, "y_dim x_dim"]:
        """Sample from the distribution.

        **Arguments:**

        - `rng_key` :
            The RNG key used to sample the noise model.
        """
        raise NotImplementedError

    @abstractmethod
    def compute_signal(
        self,
        *,
        outputs_real_space: bool = True,
        mask: AbstractMask | None = None,
        filter: AbstractFilter | None = None,
    ) -> Inexact[Array, "y_dim x_dim"]:
        """Compute the signal of the image formation model."""
        raise NotImplementedError


class AbstractLikelihoodNoiseModel(AbstractNoiseModel, strict=True):
    """A noise model equipped with a likelihood."""

    @abstractmethod
    def log_likelihood(
        self,
        observed: Inexact[Array, "y_dim x_dim"],
        *,
        mask: AbstractMask | None = None,
        filter: AbstractFilter | None = None,
    ) -> Float[Array, ""]:
        """Evaluate the log likelihood.

        **Arguments:**

        - `observed` : The observed data.
        """
        raise NotImplementedError


class AbstractEmpiricalNoiseModel(AbstractNoiseModel, strict=True):
    """A noise model that tries to empirically model the noise in
    the image. This class is not compatible with detector or
    solvent models.
    """

    def __check_init__(self):
        if _has_detector(self):
            cls_name = self.__class__.__name__
            raise ValueError(
                "Found an `AbstractDetector` class when instantiating "
                f"{cls_name}, but cryoJAX `AbstractEmpiricalNoiseModel`s are "
                "not compatible with detector classes."
            )
        if _has_solvent(self):
            cls_name = self.__class__.__name__
            raise ValueError(
                "Found an `AbstractRandomSolvent2D` class when instantiating "
                f"{cls_name}, but cryoJAX `AbstractEmpiricalNoiseModel`s are "
                "not compatible with solvent classes."
            )


def _has_detector(pytree) -> bool:
    is_leaf = lambda x: isinstance(x, AbstractDetector)
    boolean_pytree = jax.tree.map(is_leaf, pytree, is_leaf=is_leaf)
    return jax.tree.reduce(lambda x, y: x or y, boolean_pytree)


def _has_solvent(pytree) -> bool:
    is_leaf = lambda x: isinstance(x, AbstractRandomSolvent2D)
    boolean_pytree = jax.tree.map(is_leaf, pytree, is_leaf=is_leaf)
    return jax.tree.reduce(lambda x, y: x or y, boolean_pytree)



"""
Abstraction of rotations represented by matrix lie groups.
"""

from abc import abstractmethod
from collections.abc import Sequence
from typing import ClassVar
from typing_extensions import Self, override

import equinox as eqx
import jax
import jax.numpy as jnp
from equinox import AbstractClassVar
from jaxtyping import Array, Float, PRNGKeyArray

from ..jax_util import FloatLike, NDArrayLike
from ._rotation import AbstractRotation


class AbstractMatrixLieGroup(AbstractRotation, strict=True):
    """Base class for a rotation that is represented by
    a matrix lie group.

    The class is almost exactly derived from the `jaxlie.MatrixLieGroup`
    object.

    `jaxlie` was written for [Yi, Brent, et al. 2021](https://ieeexplore.ieee.org/abstract/document/9636300).
    """

    parameter_dimension: AbstractClassVar[int]
    tangent_dimension: AbstractClassVar[int]
    matrix_dimension: AbstractClassVar[int]

    def __check_init__(self):
        if self.parameter_dimension != self.parameters.size:
            cls = self.__class__.__name__
            raise AttributeError(
                "Found incorrect array size in "
                f"in class {cls}. Expected "
                f"an array with size {self.parameter_dimension} as input, "
                f"but got an array of size {self.parameters.size}."
            )

    @property
    @abstractmethod
    def parameters(self) -> Array:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def exp(cls, tangent: Array) -> Self:
        """Computes the exponential map of an element of the
        lie algebra.
        """
        raise NotImplementedError

    @abstractmethod
    def log(self) -> Array:
        """Computes the logarithmic map of the lie group element."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_matrix(cls, matrix: Array) -> Self:
        """Computes the group element from a rotation matrix."""
        raise NotImplementedError

    @abstractmethod
    def as_matrix(self) -> Array:
        """Represent the group element as a rotation matrix."""
        raise NotImplementedError

    @abstractmethod
    def normalize(self) -> Self:
        """Projects onto a group element."""
        raise NotImplementedError


class SO3(AbstractMatrixLieGroup, strict=True):
    """A rotation in 3D space, represented by the
    SO3 matrix lie group.

    The class is almost exactly derived from `jaxlie.SO3`.

    `jaxlie` was written for [Yi, Brent, et al. 2021](https://ieeexplore.ieee.org/abstract/document/9636300).
    """

    space_dimension: ClassVar[int] = 3
    parameter_dimension: ClassVar[int] = 4
    tangent_dimension: ClassVar[int] = 3
    matrix_dimension: ClassVar[int] = 3

    wxyz: Float[Array, "4"]

    def __init__(self, wxyz: Float[NDArrayLike, "4"] | Sequence[float]):
        """**Arguments:**

        - `wxyz`:
            A quaternion represented as $(q_w, q_x, q_y, q_z)$.
        """
        self.wxyz = jnp.asarray(wxyz, dtype=float)

    @property
    @override
    def parameters(self) -> Float[Array, "4"]:
        return self.wxyz

    @override
    def apply(self, target: Float[Array, "3"]) -> Float[Array, "3"]:
        # Compute using quaternion multiplys.
        padded_target = jnp.concatenate([jnp.zeros(1), target])
        return (self @ SO3(wxyz=padded_target) @ self.inverse()).wxyz[1:]

    @override
    def compose(self, other: Self) -> Self:
        w0, x0, y0, z0 = self.wxyz
        w1, x1, y1, z1 = other.wxyz
        cls = type(self)
        return cls(
            wxyz=jnp.array(
                [
                    -x0 * x1 - y0 * y1 - z0 * z1 + w0 * w1,
                    x0 * w1 + y0 * z1 - z0 * y1 + w0 * x1,
                    -x0 * z1 + y0 * w1 + z0 * x1 + w0 * y1,
                    x0 * y1 - y0 * x1 + z0 * w1 + w0 * z1,
                ]
            )
        )

    @override
    def inverse(self) -> Self:
        # Negate complex terms.
        return eqx.tree_at(lambda R: R.wxyz, self, self.wxyz * jnp.array([1, -1, -1, -1]))

    @classmethod
    def from_x_radians(cls, angle: FloatLike) -> Self:
        """Generates a x-axis rotation."""
        return cls.exp(jnp.asarray([angle, 0.0, 0.0]))

    @classmethod
    def from_y_radians(cls, angle: FloatLike) -> Self:
        """Generates a x-axis rotation."""
        return cls.exp(jnp.asarray([0.0, angle, 0.0]))

    @classmethod
    def from_z_radians(cls, angle: FloatLike) -> Self:
        """Generates a x-axis rotation."""
        return cls.exp(jnp.asarray([0.0, 0.0, angle]))

    @override
    @classmethod
    def identity(cls) -> Self:
        return cls(jnp.asarray([1.0, 0.0, 0.0, 0.0]))

    @override
    @classmethod
    def from_matrix(cls, matrix: Float[Array, "3 3"]) -> Self:
        # Modified from:
        # > "Converting a Rotation Matrix to a Quaternion" from Mike Day
        # > https://d3cw3dd2w32x2b.cloudfront.net/wp-content/uploads/2015/01/matrix-to-quat.pdf

        def case0(m):
            t = 1 + m[0, 0] - m[1, 1] - m[2, 2]
            q = jnp.array(
                [
                    m[2, 1] - m[1, 2],
                    t,
                    m[1, 0] + m[0, 1],
                    m[0, 2] + m[2, 0],
                ]
            )
            return t, q

        def case1(m):
            t = 1 - m[0, 0] + m[1, 1] - m[2, 2]
            q = jnp.array(
                [
                    m[0, 2] - m[2, 0],
                    m[1, 0] + m[0, 1],
                    t,
                    m[2, 1] + m[1, 2],
                ]
            )
            return t, q

        def case2(m):
            t = 1 - m[0, 0] - m[1, 1] + m[2, 2]
            q = jnp.array(
                [
                    m[1, 0] - m[0, 1],
                    m[0, 2] + m[2, 0],
                    m[2, 1] + m[1, 2],
                    t,
                ]
            )
            return t, q

        def case3(m):
            t = 1 + m[0, 0] + m[1, 1] + m[2, 2]
            q = jnp.array(
                [
                    t,
                    m[2, 1] - m[1, 2],
                    m[0, 2] - m[2, 0],
                    m[1, 0] - m[0, 1],
                ]
            )
            return t, q

        # Compute four cases, then pick the most precise one.
        # Probably worth revisiting this!
        case0_t, case0_q = case0(matrix)
        case1_t, case1_q = case1(matrix)
        case2_t, case2_q = case2(matrix)
        case3_t, case3_q = case3(matrix)

        cond0 = matrix[2, 2] < 0
        cond1 = matrix[0, 0] > matrix[1, 1]
        cond2 = matrix[0, 0] < -matrix[1, 1]

        t = jnp.where(
            cond0,
            jnp.where(cond1, case0_t, case1_t),
            jnp.where(cond2, case2_t, case3_t),
        )
        q = jnp.where(
            cond0,
            jnp.where(cond1, case0_q, case1_q),
            jnp.where(cond2, case2_q, case3_q),
        )

        return cls(wxyz=q * 0.5 / jnp.sqrt(t))

    @override
    def as_matrix(self) -> Float[Array, "3 3"]:
        norm = self.wxyz @ self.wxyz
        q = self.wxyz * jnp.sqrt(2.0 / norm)
        q = jnp.outer(q, q)
        return jnp.array(
            [
                [1.0 - q[2, 2] - q[3, 3], q[1, 2] - q[3, 0], q[1, 3] + q[2, 0]],
                [q[1, 2] + q[3, 0], 1.0 - q[1, 1] - q[3, 3], q[2, 3] - q[1, 0]],
                [q[1, 3] - q[2, 0], q[2, 3] + q[1, 0], 1.0 - q[1, 1] - q[2, 2]],
            ]
        )

    @classmethod
    def exp(cls, tangent: Float[Array, "3"]) -> Self:
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/so3.hpp#L583
        theta_squared = tangent @ tangent
        theta_pow_4 = theta_squared * theta_squared
        use_taylor = theta_squared < _get_epsilon(tangent.dtype)

        # Shim to avoid NaNs in jnp.where branches, which cause failures for
        # reverse-mode AD.
        safe_theta = jnp.sqrt(
            jnp.where(
                use_taylor,
                1.0,  # Any constant value should do here.
                theta_squared,
            )
        )
        safe_half_theta = 0.5 * safe_theta

        real_factor = jnp.where(
            use_taylor,
            1.0 - theta_squared / 8.0 + theta_pow_4 / 384.0,
            jnp.cos(safe_half_theta),
        )

        imaginary_factor = jnp.where(
            use_taylor,
            0.5 - theta_squared / 48.0 + theta_pow_4 / 3840.0,
            jnp.sin(safe_half_theta) / safe_theta,
        )

        return cls(
            wxyz=jnp.concatenate(
                [
                    real_factor[None],
                    imaginary_factor * tangent,
                ]
            )
        )

    @override
    def log(self) -> Float[Array, "3"]:
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/so3.hpp#L247

        w = self.wxyz[..., 0]
        norm_sq = self.wxyz[..., 1:] @ self.wxyz[..., 1:]
        use_taylor = norm_sq < _get_epsilon(norm_sq.dtype)

        # Shim to avoid NaNs in jnp.where branches, which cause failures for
        # reverse-mode AD.
        norm_safe = jnp.sqrt(
            jnp.where(
                use_taylor,
                1.0,  # Any non-zero value should do here.
                norm_sq,
            )
        )
        w_safe = jnp.where(use_taylor, w, 1.0)
        atan_n_over_w = jnp.arctan2(
            jnp.where(w < 0, -norm_safe, norm_safe),
            jnp.abs(w),
        )
        atan_factor = jnp.where(
            use_taylor,
            2.0 / w_safe - 2.0 / 3.0 * norm_sq / w_safe**3,
            jnp.where(
                jnp.abs(w) < _get_epsilon(w.dtype),
                jnp.where(w > 0, 1.0, -1.0) * jnp.pi / norm_safe,
                2.0 * atan_n_over_w / norm_safe,
            ),
        )

        return atan_factor * self.wxyz[1:]

    @override
    def normalize(self) -> Self:
        return eqx.tree_at(lambda R: R.wxyz, self, self.wxyz / jnp.linalg.norm(self.wxyz))

    @classmethod
    def sample_uniform(cls, key: PRNGKeyArray) -> Self:
        # Uniformly sample over S^3.
        # > Reference: http://planning.cs.uiuc.edu/node198.html
        u1, u2, u3 = jax.random.uniform(
            key=key,
            shape=(3,),
            minval=jnp.zeros(3),
            maxval=jnp.array([1.0, 2.0 * jnp.pi, 2.0 * jnp.pi]),
        )
        a = jnp.sqrt(1.0 - u1)
        b = jnp.sqrt(u1)

        return cls(
            wxyz=jnp.array(
                [
                    a * jnp.sin(u2),
                    a * jnp.cos(u2),
                    b * jnp.sin(u3),
                    b * jnp.cos(u3),
                ]
            )
        )


class SO2(AbstractMatrixLieGroup, strict=True):
    """A rotation in 2D space, represented by the
    SO2 matrix lie group.

    The class is almost exactly derived from `jaxlie.SO2`.

    `jaxlie` was written for [Yi, Brent, et al. 2021](https://ieeexplore.ieee.org/abstract/document/9636300).
    """

    space_dimension: ClassVar[int] = 2
    parameter_dimension: ClassVar[int] = 2
    tangent_dimension: ClassVar[int] = 1
    matrix_dimension: ClassVar[int] = 2

    unit_complex: Float[Array, "2"]

    def __init__(
        self, unit_complex: Float[NDArrayLike, "2"] | Sequence[float | NDArrayLike]
    ):
        r"""**Arguments:**

        - `unit_complex`:
            A complex number represented as a 2-vector
            $(cos \theta, sin \theta)$.
        """
        self.unit_complex = jnp.asarray(unit_complex, dtype=float)

    @property
    @override
    def parameters(self) -> Float[Array, "4"]:
        return self.unit_complex

    @classmethod
    def from_radians(cls, theta: FloatLike) -> Self:
        """Construct a rotation object from a scalar angle."""
        cos = jnp.cos(theta)
        sin = jnp.sin(theta)
        return cls(unit_complex=jnp.stack([cos, sin], axis=-1))

    def as_radians(self) -> Float[Array, ""]:
        """Compute a scalar angle from a rotation object."""
        radians = self.log()
        return radians

    @override
    @classmethod
    def identity(cls) -> Self:
        return cls(jnp.asarray([1.0, 0.0]))

    @override
    @classmethod
    def from_matrix(cls, matrix: Float[NDArrayLike, "2 2"]) -> Self:
        return cls(unit_complex=jnp.asarray(matrix[:, 0]))

    @override
    def as_matrix(self) -> Float[Array, "2 2"]:
        cos_sin = self.unit_complex
        out = jnp.stack(
            [
                # [cos, -sin],
                cos_sin * jnp.array([1, -1]),
                # [sin, cos],
                cos_sin[::-1],
            ],
            axis=-1,
        )
        return out

    @override
    def apply(self, target: Float[Array, "2"]) -> Float[Array, "2"]:
        return self.as_matrix() @ target

    @override
    def compose(self, other: Self) -> Self:
        cls = type(self)
        return cls(unit_complex=self.as_matrix() @ other.unit_complex)

    @classmethod
    @override
    def exp(cls, tangent: FloatLike) -> Self:
        cos = jnp.cos(tangent)
        sin = jnp.sin(tangent)
        return cls(unit_complex=jnp.concatenate([cos, sin], axis=-1))

    @override
    def log(self) -> Float[Array, ""]:
        return jnp.arctan2(self.unit_complex[1], self.unit_complex[0])

    @override
    def inverse(self) -> Self:
        cls = type(self)
        return cls(unit_complex=self.unit_complex * jnp.array([1, -1]))

    @override
    def normalize(self) -> Self:
        cls = type(self)
        return cls(
            unit_complex=self.unit_complex / jnp.linalg.norm(self.unit_complex, axis=-1)
        )

    @override
    @classmethod
    def sample_uniform(cls, key: PRNGKeyArray) -> Self:
        return cls.from_radians(
            jax.random.uniform(key=key, shape=(), minval=0.0, maxval=2.0 * jnp.pi)
        )


def _get_epsilon(dtype: jnp.dtype) -> float:
    """Helper for grabbing type-specific precision constants."""
    return {
        jnp.dtype("float32"): 1e-5,
        jnp.dtype("float64"): 1e-10,
    }[dtype]



"""
Real voxel-based representations of a volume.
"""

import math
from typing import Any, ClassVar, Literal, cast
from typing_extensions import Self, override

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Float

from ...jax_util import NDArrayLike
from ...ndimage import convert_fftn_to_rfftn, crop_to_shape, irfftn, make_coordinate_grid
from .._image_config import AbstractImageConfig
from .._pose import AbstractPose
from .base_volume import AbstractVolumeIntegrator, AbstractVoxelVolume, ProjectionArray


try:
    import jax_finufft as jnufft

    JAX_FINUFFT_IMPORT_ERROR = None
except ModuleNotFoundError as err:
    jnufft = None
    JAX_FINUFFT_IMPORT_ERROR = err


class AbstractRealVoxelVolume(AbstractVoxelVolume, strict=True):
    """Abstract interface for a voxel-based volume."""

    coordinate_grid_in_pixels: eqx.AbstractVar[Float[Array, "dim dim dim 3"]]

    @override
    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Return a new volume with a rotated
        `coordinate_grid_in_pixels`.
        """
        return eqx.tree_at(
            lambda d: d.coordinate_grid_in_pixels,
            self,
            pose.rotate_coordinates(self.coordinate_grid_in_pixels, inverse=inverse),
        )


class RealVoxelGridVolume(AbstractRealVoxelVolume, strict=True):
    """A 3D voxel grid in real-space."""

    real_voxel_grid: Float[Array, "dim dim dim"]
    coordinate_grid_in_pixels: Float[Array, "dim dim dim 3"]

    rotation_convention: ClassVar[Literal["frame"]] = "frame"

    def __init__(
        self,
        real_voxel_grid: Float[NDArrayLike, "dim dim dim"],
        coordinate_grid_in_pixels: Float[NDArrayLike, "dim dim dim 3"],
    ):
        """**Arguments:**

        - `real_voxel_grid`: The voxel grid in fourier space.
        - `coordinate_grid_in_pixels`: A coordinate grid.
        """
        self.real_voxel_grid = jnp.asarray(real_voxel_grid, dtype=float)
        self.coordinate_grid_in_pixels = jnp.asarray(
            coordinate_grid_in_pixels, dtype=float
        )

    @property
    def shape(self) -> tuple[int, int, int]:
        """The shape of the `real_voxel_grid`."""
        return cast(tuple[int, int, int], self.real_voxel_grid.shape)

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[NDArrayLike, "dim dim dim"],
        *,
        coordinate_grid_in_pixels: Float[Array, "dim dim dim 3"] | None = None,
        crop_scale: float | None = None,
    ) -> Self:
        """Load a `RealVoxelGridVolume` from a real-valued 3D voxel grid.

        **Arguments:**

        - `real_voxel_grid`: A voxel grid in real space.
        - `crop_scale`: Scale factor at which to crop `real_voxel_grid`.
                        Must be a value greater than `1`.
        """
        # Cast to jax array
        real_voxel_grid = jnp.asarray(real_voxel_grid, dtype=float)
        # Make coordinates if not given
        if coordinate_grid_in_pixels is None:
            # Option for cropping template
            if crop_scale is not None:
                if crop_scale < 1.0:
                    raise ValueError("`crop_scale` must be greater than 1.0")
                cropped_shape = cast(
                    tuple[int, int, int],
                    tuple([int(s / crop_scale) for s in real_voxel_grid.shape[-3:]]),
                )
                real_voxel_grid = crop_to_shape(real_voxel_grid, cropped_shape)
            coordinate_grid_in_pixels = make_coordinate_grid(real_voxel_grid.shape[-3:])

        return cls(real_voxel_grid, coordinate_grid_in_pixels)


class RealVoxelProjection(
    AbstractVolumeIntegrator[RealVoxelGridVolume],
    strict=True,
):
    """Integrate points onto the exit plane using non-uniform FFTs."""

    eps: float
    opts: Any

    outputs_ewald_sphere: ClassVar[bool] = False

    def __init__(self, *, eps: float = 1e-6, opts: Any = None):
        """**Arguments:**

        - `eps`:
            See [`jax-finufft`](https://github.com/flatironinstitute/jax-finufft)
            for documentation.
        - `opts`:
            A `jax_finufft.options.Opts` or `jax_finufft.options.NestedOpts`
            dataclass.
            See [`jax-finufft`](https://github.com/flatironinstitute/jax-finufft)
            for documentation.
        """
        if jnufft is None:
            raise RuntimeError(
                "Tried to use the `RealVoxelProjection` "
                "class, but `jax-finufft` is not installed. "
                "See https://github.com/flatironinstitute/jax-finufft "
                "for installation instructions."
            ) from JAX_FINUFFT_IMPORT_ERROR
        self.eps = eps
        self.opts = opts

    @override
    def integrate(
        self,
        volume_representation: RealVoxelGridVolume,
        image_config: AbstractImageConfig,
        outputs_real_space: bool = False,
    ) -> ProjectionArray:
        """Integrate the volume at the `AbstractImageConfig` settings
        of a voxel-based representation in real-space, using non-uniform FFTs.

        **Arguments:**

        - `volume_representation`:
            The volume representation.
        - `image_config`:
            The image configuration.
        - `outputs_real_space`:
            If `True`, return the image in real space. Otherwise,
            return in Fourier.

        **Returns:**

        The volume projection in real or Fourier space, at the
        `image_config.padded_shape`.
        """
        n_voxels = math.prod(volume_representation.shape)
        fourier_projection = _project_with_nufft(
            volume_representation.real_voxel_grid.ravel(),
            volume_representation.coordinate_grid_in_pixels.reshape((n_voxels, 3)),
            image_config.padded_shape,
            eps=self.eps,
            opts=self.opts,
        )
        # Scale by voxel size for units
        fourier_projection *= image_config.pixel_size
        return (
            irfftn(fourier_projection, s=image_config.padded_shape)
            if outputs_real_space
            else fourier_projection
        )


def _project_with_nufft(weights, coordinate_list, shape, eps=1e-6, opts=None):
    assert jnufft is not None
    weights, coordinate_list = (
        jnp.asarray(weights, dtype=complex),
        jnp.asarray(coordinate_list, dtype=float),
    )
    # Get x and y coordinates
    coordinates_xy = coordinate_list[:, :2]
    # Normalize coordinates betweeen -pi and pi
    ny, nx = shape
    box_xy = jnp.asarray((nx, ny))
    coordinates_periodic = 2 * jnp.pi * coordinates_xy / box_xy
    # Unpack and compute
    x, y = coordinates_periodic[:, 0], coordinates_periodic[:, 1]
    fourier_projection = jnufft.nufft1(shape, weights, y, x, eps=eps, opts=opts, iflag=-1)
    # Shift zero frequency component to corner
    fourier_projection = jnp.fft.ifftshift(fourier_projection)
    # Convert to rfftn output
    return convert_fftn_to_rfftn(fourier_projection, mode="real")



import cryojax.simulator as cxs
import equinox as eqx
import pytest


@pytest.mark.parametrize(
    "padded_shape, precompute_mode",
    (
        ((5, 5), "rfft"),
        ((10, 10), "rfft"),
        ((5, 5), "fft"),
        ((10, 10), "fft"),
        ((5, 5), "all"),
        ((10, 10), "all"),
    ),
)
def test_precompute(padded_shape, precompute_mode):
    c = cxs.BasicImageConfig(
        (5, 5),
        pixel_size=1.0,
        voltage_in_kilovolts=300.0,
        pad_options=dict(shape=padded_shape),
        precompute_mode=precompute_mode,
    )
    precomputed_grids = c.precomputed_grids
    assert precomputed_grids is not None
    # rfftfreq grids
    assert c.get_frequency_grid(padding=False, physical=False) is precomputed_grids.get(
        real_space=False, padding=False
    )
    assert c.get_frequency_grid(padding=True, physical=False) is precomputed_grids.get(
        real_space=False, padding=True
    )
    # coordinate grids
    if precompute_mode == "all":
        assert c.get_coordinate_grid(
            padding=False, physical=False
        ) is precomputed_grids.get(real_space=True, padding=False)
        assert c.get_coordinate_grid(
            padding=True, physical=False
        ) is precomputed_grids.get(real_space=True, padding=True)
    else:
        with pytest.raises(Exception):
            precomputed_grids.get(real_space=True)
        with pytest.raises(Exception):
            precomputed_grids.get(real_space=True, padding=True)
    # fftfreq grids
    if precompute_mode == "rfft":
        with pytest.raises(Exception):
            precomputed_grids.get(real_space=False, full=True)
        with pytest.raises(Exception):
            precomputed_grids.get(real_space=False, full=True, padding=True)
    else:
        assert c.get_frequency_grid(
            padding=True, physical=False, full=True
        ) is precomputed_grids.get(real_space=False, padding=True, full=True)
        assert c.get_frequency_grid(
            padding=False, physical=False, full=True
        ) is precomputed_grids.get(real_space=False, padding=False, full=True)


def test_compile_time_eval():
    c = cxs.BasicImageConfig(
        (5, 5),
        pixel_size=1.0,
        voltage_in_kilovolts=300.0,
        pad_options=dict(shape=(10, 10)),
        precompute_mode="compile_time_eval",
    )

    @eqx.filter_jit
    def _get_coords(_c):
        return _c.get_coordinate_grid()

    @eqx.filter_jit
    def _get_freqs(_c):
        return _c.get_coordinate_grid()

    _ = _get_coords(c)
    _ = _get_freqs(c)



"""
Check coverage with
pytest --cov-report term-missing:skip-covered --cov=src/cryojax/data/_relion tests/test_relion_dataset.py
"""  # noqa

import os
import shutil
from functools import partial
from typing import cast

import cryojax.ndimage as im
import cryojax.simulator as cxs
import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jr
import numpy as np
import pandas as pd
import pytest
from cryojax.dataset import (
    RelionParticleDataset,
    RelionParticleParameterFile,
    simulate_particle_stack,
)
from cryojax.dataset._particle_data.relion import (
    _validate_starfile_data,
)
from cryojax.io import read_array_from_mrc
from cryojax.rotations import SO3
from jaxtyping import TypeCheckError


def compare_pytrees(pytree1, pytree2):
    del pytree1, pytree2
    return True
    # arrays1, others1 = eqx.partition(pytree1, eqx.is_array)
    # arrays2, others2 = eqx.partition(pytree2, eqx.is_array)

    # bool_arrays = all(
    #     jax.tree.leaves(jax.tree.map(lambda x, y: jnp.allclose(x, y), arrays1, arrays2))
    # )
    # bool_others = all(
    #     jax.tree.leaves(jax.tree.map(lambda x, y: x == y, others1, others2))
    # )
    # return bool_arrays and bool_others


@pytest.fixture
def sample_starfile_path():
    return os.path.join(os.path.dirname(__file__), "data", "test_starfile.star")


@pytest.fixture
def sample_relion_project_path():
    return os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture
def sample_image_stack_path(sample_relion_project_path):
    return os.path.join(sample_relion_project_path, "000000.mrcs")


@pytest.fixture
def sample_image_stack(sample_image_stack_path):
    return read_array_from_mrc(sample_image_stack_path)


@pytest.fixture
def parameter_file(sample_starfile_path):
    return RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        loads_envelope=True,
        loads_metadata=True,
    )


@pytest.fixture
def relion_parameters():
    image_config = cxs.BasicImageConfig(
        shape=(4, 4),
        pixel_size=1.5,
        voltage_in_kilovolts=300.0,
        pad_options=dict(mode="constant", shape=(14, 14)),
    )

    pose = cxs.EulerAnglePose()
    transfer_theory = cxs.ContrastTransferTheory(
        ctf=cxs.AstigmaticCTF(),
    )
    return dict(image_config=image_config, pose=pose, transfer_theory=transfer_theory)


#
# Tests for starfile loading


class TestErrorRaisingForLoading:
    def test_load_with_badparticle_name(self, parameter_file, sample_relion_project_path):
        parameter_file.starfile_data["particles"].loc[0, "rlnImageName"] = "0.0"
        dataset = RelionParticleDataset(
            path_to_relion_project=sample_relion_project_path,
            parameter_file=parameter_file,
        )
        with pytest.raises(ValueError):
            dataset[0]

        def test_load_with_badparticle_name2(
            self, parameter_file, sample_relion_project_path
        ):
            parameter_file.starfile_data["particles"].loc[0, "rlnImageName"] = "0000.mrcs"
            dataset = RelionParticleDataset(
                path_to_relion_project=sample_relion_project_path,
                parameter_file=parameter_file,
            )
            with pytest.raises(TypeError):
                dataset[0]

    def test_load_with_bad_shape(self, parameter_file, sample_relion_project_path):
        parameter_file.starfile_data["optics"].loc[0, "rlnImageSize"] = 1
        dataset = RelionParticleDataset(
            path_to_relion_project=sample_relion_project_path,
            parameter_file=parameter_file,
        )
        with pytest.raises(ValueError):
            dataset[0]

    def test_with_bad_indices(self, parameter_file, sample_relion_project_path):
        dataset = RelionParticleDataset(
            path_to_relion_project=sample_relion_project_path,
            parameter_file=parameter_file,
        )

        # overflow index
        with pytest.raises(IndexError):
            parameter_file[len(parameter_file)]

        with pytest.raises(IndexError):
            dataset[len(dataset)]

        # overflow slice
        with pytest.raises(IndexError):
            parameter_file[len(parameter_file) :]

        with pytest.raises(IndexError):
            dataset[len(dataset) :]

        # wrong index type
        with pytest.raises(IndexError):
            parameter_file["wrong_index"]

        with pytest.raises(IndexError):
            dataset["wrong_index"]  # type: ignore

    def test_validate_starfile_data(self):
        with pytest.raises(ValueError):
            _validate_starfile_data({"wrong": pd.DataFrame({})})

        with pytest.raises(ValueError):
            _validate_starfile_data({"particles": pd.DataFrame({})})

        mock_particles_df = pd.DataFrame(
            {
                "rlnDefocusU": 0.0,
                "rlnDefocusV": 0.0,
                "rlnDefocusAngle": 0.0,
                "rlnPhaseShift": 0.0,
                "rlnImageName": "mock.mrcs",
            },
            index=[0],
        )
        with pytest.raises(ValueError):
            _validate_starfile_data({"particles": mock_particles_df})

        with pytest.raises(ValueError):
            _validate_starfile_data(
                {"particles": mock_particles_df, "optics": pd.DataFrame({})}
            )


def test_pad_options(sample_starfile_path):
    """Test the pad_options argument to the parameter file."""
    # Test with a valid input

    pad_options = dict(shape=(22, 22))
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        loads_envelope=True,
        loads_metadata=True,
        pad_options=pad_options,
    )
    image_config = parameter_file[0]["image_config"]

    ref_config = cxs.BasicImageConfig(
        shape=(16, 16),
        pixel_size=12.0,
        voltage_in_kilovolts=300.0,
        pad_options=pad_options,
    )
    assert image_config.shape == ref_config.shape
    assert image_config.pixel_size == np.asarray(ref_config.pixel_size)
    assert image_config.voltage_in_kilovolts == np.asarray(
        ref_config.voltage_in_kilovolts
    )

    assert image_config.padded_shape == ref_config.padded_shape


def test_load_starfile_envelope_params(sample_starfile_path):
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        loads_envelope=True,
        loads_metadata=False,
    )

    assert parameter_file.loads_envelope is True
    parameters = parameter_file[0]
    assert parameters["transfer_theory"].envelope is not None

    parameters = parameter_file[:]
    assert parameters["transfer_theory"].envelope is not None

    envelope = parameters["transfer_theory"].envelope

    particle_data = parameter_file.starfile_data["particles"]
    # check that envelope params match
    for i in range(len(parameter_file)):
        # check b-factors
        np.testing.assert_allclose(
            envelope.b_factor[i],  # type: ignore
            particle_data["rlnCtfBfactor"][i],
            rtol=1e-5,
        )
        np.testing.assert_allclose(
            envelope.amplitude[i],  # type: ignore
            particle_data["rlnCtfScalefactor"][i],
            rtol=1e-5,
        )
    return


def test_load_starfile_ctf_params(sample_starfile_path):
    is_shape = lambda shape, pytree: jax.tree.reduce(
        lambda x, y: x and y,
        jax.tree.map(lambda x: x.shape == shape, pytree),
    )

    def compute_defocus(defU, defV):
        return 0.5 * (defU + defV)

    def compute_astigmatism(defU, defV):
        return defU - defV

    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        loads_envelope=False,
        loads_metadata=True,
    )

    assert parameter_file.loads_envelope is False

    # Check loading 1 particle
    parameters = parameter_file[0]
    assert parameters["transfer_theory"].envelope is None
    assert is_shape((), eqx.filter(parameters["transfer_theory"].ctf, eqx.is_array))

    # Check loading >1 particles
    parameters = parameter_file[:]
    assert parameters["transfer_theory"].envelope is None
    assert is_shape(
        (len(parameter_file),),
        eqx.filter(parameters["transfer_theory"].ctf, eqx.is_array),
    )

    transfer_theory = parameters["transfer_theory"]
    ctf = cast(cxs.AstigmaticCTF, transfer_theory.ctf)

    particle_data = parameter_file.starfile_data["particles"]
    # check CTF parameters
    for i in range(len(parameter_file)):
        # defocus
        np.testing.assert_allclose(
            ctf.defocus_in_angstroms[i],
            compute_defocus(
                particle_data["rlnDefocusU"][i],
                particle_data["rlnDefocusV"][i],
            ),
            rtol=1e-5,
        )

        # astigmatism
        np.testing.assert_allclose(
            ctf.astigmatism_in_angstroms[i],
            compute_astigmatism(
                particle_data["rlnDefocusU"][i],
                particle_data["rlnDefocusV"][i],
            ),
            rtol=1e-5,
        )

        # astigmatism_angle
        np.testing.assert_allclose(
            ctf.astigmatism_angle[i],
            particle_data["rlnDefocusAngle"][i],
            rtol=1e-5,
        )

        # phase shift
        np.testing.assert_allclose(
            transfer_theory.phase_shift[i],
            particle_data["rlnPhaseShift"][i],
            rtol=1e-5,
        )

    return


def test_load_starfile_pose_params(sample_starfile_path):
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        loads_envelope=False,
        loads_metadata=True,
        rotation_convention="object",
    )

    parameters = parameter_file[:]
    pose = parameters["pose"]

    particle_data = parameter_file.starfile_data["particles"]
    # Check pose parameters
    for i in range(len(parameter_file)):
        # offset x
        np.testing.assert_allclose(
            pose.offset_x_in_angstroms[i],
            -particle_data["rlnOriginXAngst"][i],  # conventions!
            rtol=1e-5,
        )

        # offset y
        np.testing.assert_allclose(
            pose.offset_y_in_angstroms[i],
            -particle_data["rlnOriginYAngst"][i],  # conventions!
            rtol=1e-5,
        )

        # phi angle - AngleRot
        np.testing.assert_allclose(
            pose.phi_angle[i],
            -particle_data["rlnAngleRot"][i],
            rtol=1e-5,
        )

        # theta angle - AngleTilt
        np.testing.assert_allclose(
            pose.theta_angle[i],
            -particle_data["rlnAngleTilt"][i],
            rtol=1e-5,
        )

        # psi angle - AnglePsi
        np.testing.assert_allclose(
            pose.psi_angle[i],
            -particle_data["rlnAnglePsi"][i],
            rtol=1e-5,
        )


def test_load_starfile_pose_inverse(sample_starfile_path):
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        loads_envelope=False,
        loads_metadata=True,
        rotation_convention="object",
    )
    parameter_file_inverse = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        loads_envelope=False,
        loads_metadata=True,
        rotation_convention="frame",
    )

    # Without batch dim
    parameters = parameter_file[0]
    pose = parameters["pose"]
    parameters_inverse = parameter_file_inverse[0]
    pose_inverse = parameters_inverse["pose"]

    get_rotation = lambda p: p.rotation
    rotation = get_rotation(pose)
    rotation_inverse = get_rotation(pose_inverse)

    np.testing.assert_allclose(rotation.wxyz, rotation_inverse.wxyz.at[1:].mul(-1))

    # With batch dim
    parameters = parameter_file[:]
    pose = parameters["pose"]
    parameters_inverse = parameter_file_inverse[:]
    pose_inverse = parameters_inverse["pose"]

    get_rotation = eqx.filter_vmap(lambda p: p.rotation)
    rotation = get_rotation(pose)
    rotation_inverse = get_rotation(pose_inverse)

    np.testing.assert_allclose(rotation.wxyz, rotation_inverse.wxyz.at[:, 1:].mul(-1))


def test_load_starfile_wo_metadata(sample_starfile_path):
    """Test loading a starfile without metadata."""
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        loads_envelope=False,
        loads_metadata=False,
    )

    # check that metadata is empty dict
    assert parameter_file[0]["metadata"] is None
    assert parameter_file[:]["metadata"] is None
    assert not parameter_file.loads_metadata


def test_load_optics_group_broadcasting(sample_starfile_path):
    """Test loading a starfile with optics group."""
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        loads_envelope=False,
        loads_metadata=True,
        broadcasts_image_config=True,
    )

    parameters = parameter_file[:]
    image_config = parameters["image_config"]
    assert image_config.voltage_in_kilovolts.ndim > 0
    assert image_config.pixel_size.ndim > 0
    assert parameter_file.broadcasts_image_config is True

    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        loads_envelope=False,
        loads_metadata=True,
        broadcasts_image_config=False,
    )
    parameters = parameter_file[:]
    image_config = parameters["image_config"]
    assert image_config.voltage_in_kilovolts.ndim == 0
    assert image_config.pixel_size.ndim == 0
    assert parameter_file.broadcasts_image_config is False

    return


def test_parameter_file_setters(sample_starfile_path):
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        loads_envelope=False,
        loads_metadata=False,
        broadcasts_image_config=False,
        updates_optics_group=False,
    )

    parameter_file.loads_metadata = True
    assert parameter_file.loads_metadata

    parameter_file.loads_envelope = True
    assert parameter_file.loads_envelope

    parameter_file.broadcasts_image_config = True
    assert parameter_file.broadcasts_image_config

    parameter_file.updates_optics_group = True
    assert parameter_file.updates_optics_group


def test_load_starfile_vs_mrcs_shape(sample_starfile_path, sample_relion_project_path):
    """Test loading a starfile with mrcs."""
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        loads_envelope=False,
        loads_metadata=False,
        broadcasts_image_config=False,
    )
    dataset = RelionParticleDataset(
        parameter_file, sample_relion_project_path, loads_parameters=True
    )

    particle_stack = dataset[:]
    parameters = particle_stack["parameters"]
    assert parameters is not None
    image_config = parameters["image_config"]
    assert particle_stack["images"].shape == (
        len(parameter_file),
        *image_config.shape,
    )

    particle_stack = dataset[0]
    image_config = parameters["image_config"]
    assert particle_stack["images"].shape == image_config.shape

    particle_stack = dataset[0:2]
    image_config = parameters["image_config"]
    assert particle_stack["images"].shape == (2, *image_config.shape)

    assert len(dataset) == len(parameter_file)

    return


def test_no_load_parameters(sample_starfile_path, sample_relion_project_path):
    """Test loading a starfile with mrcs."""
    parameter_file = RelionParticleParameterFile(path_to_starfile=sample_starfile_path)
    dataset = RelionParticleDataset(
        parameter_file, sample_relion_project_path, loads_parameters=True
    )

    # For particle stack with leading dim
    dataset.loads_parameters = True
    particle_stack_params = dataset[:]
    dataset.loads_parameters = False
    particle_stack_noparams = dataset[:]

    assert particle_stack_noparams["parameters"] is None
    assert (
        particle_stack_params["images"].shape == particle_stack_noparams["images"].shape
    )
    assert isinstance(particle_stack_params["images"], np.ndarray)

    # For particle stack with no leading dim
    dataset.loads_parameters = True
    particle_stack_params = dataset[0]
    dataset.loads_parameters = False
    particle_stack_noparams = dataset[0]
    assert (
        particle_stack_params["images"].shape == particle_stack_noparams["images"].shape
    )
    assert isinstance(particle_stack_noparams["images"], np.ndarray)

    return


#
# Tests for starfile writing
#


@pytest.mark.parametrize(
    "index, loads_envelope",
    [
        (0, False),
        ([0, 1], False),
        (0, True),
    ],
)
def test_append_particle_parameters(index, loads_envelope):
    index = np.asarray(index)
    ndim = index.ndim

    @eqx.filter_vmap
    def make_particle_params(dummy_idx):
        image_config = cxs.BasicImageConfig(
            shape=(4, 4),
            pixel_size=1.5,
            voltage_in_kilovolts=300.0,
        )

        pose = cxs.EulerAnglePose()
        transfer_theory = cxs.ContrastTransferTheory(
            ctf=cxs.AstigmaticCTF(),
            envelope=im.FourierGaussian() if loads_envelope else None,
        )
        return dict(
            image_config=image_config,
            pose=pose,
            transfer_theory=transfer_theory,
        )

    # Make particle parameters, using custom metadata
    metadata = pd.DataFrame(
        data={
            "rlnMicrographName": list("dummy/micrograph.mrc" for _ in range(index.size)),
            "rlnCoordinateX": np.atleast_1d(np.full_like(index, 2, dtype=int)),
            "rlnCoordinateY": np.atleast_1d(np.full_like(index, 1, dtype=int)),
        },
    )
    particle_params = make_particle_params(jnp.atleast_1d(index))
    particle_params["metadata"] = metadata
    # ... custom metadata
    if ndim == 0:
        particle_params = jax.tree.map(
            lambda x: jnp.squeeze(x) if isinstance(x, jax.Array) else x, particle_params
        )
    # Add to dataset
    path_to_starfile = "tests/outputs/starfile_writing/test_particle_parameters.star"
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=path_to_starfile,
        mode="w",
        exists_ok=True,
        loads_envelope=loads_envelope,
        loads_metadata=False,
    )
    parameter_file.append(particle_params)
    # Make sure custom metadata was added
    particle_dataframe = parameter_file.starfile_data["particles"]
    assert set(metadata.columns).issubset(particle_dataframe.columns)
    # Make sure dataframes are the same
    metadata_extracted = particle_dataframe.loc[
        particle_dataframe.index[np.atleast_1d(index)],
        ["rlnMicrographName", "rlnCoordinateX", "rlnCoordinateY"],
    ]
    np.testing.assert_equal(metadata.to_numpy(), metadata_extracted.to_numpy())
    # Make sure parameters read and the same as what was appended
    loaded_particle_params = parameter_file[index]
    particle_params["metadata"] = None  # need to remove dataframe
    assert compare_pytrees(loaded_particle_params, particle_params)


@pytest.mark.parametrize(
    "index, updates_optics_group, sets_envelope",
    [
        (0, False, False),
        ([0, 1], False, False),
        (0, True, False),
        (0, False, True),
    ],
)
def test_set_particle_parameters(
    sample_starfile_path,
    index,
    updates_optics_group,
    sets_envelope,
):
    index = np.asarray(index)
    n_particles, ndim = index.size, index.ndim

    def make_params(rng_key) -> dict:
        rng_keys = jr.split(rng_key, n_particles)
        make_pose = eqx.filter_vmap(
            lambda rng_key: cxs.EulerAnglePose.from_rotation(SO3.sample_uniform(rng_key))
        )
        pose = make_pose(rng_keys)
        return dict(
            image_config=cxs.BasicImageConfig(
                shape=(4, 4), pixel_size=3.324, voltage_in_kilovolts=121.3
            ),
            pose=pose,
            transfer_theory=cxs.ContrastTransferTheory(
                cxs.AstigmaticCTF(defocus_in_angstroms=1234.0),
                amplitude_contrast_ratio=0.1234,
                envelope=im.FourierGaussian(b_factor=12.34) if sets_envelope else None,
            ),
        )

    metadata = pd.DataFrame(
        data={
            "rlnMicrographName": list("dummy/micrograph.mrc" for _ in range(index.size)),
            "rlnCoordinateX": np.atleast_1d(np.full_like(index, 2, dtype=int)),
            "rlnCoordinateY": np.atleast_1d(np.full_like(index, 1, dtype=int)),
        },
    )
    rng_key = jr.key(0)
    new_parameters = make_params(rng_key)
    if ndim == 0:
        new_parameters = jax.tree.map(
            lambda x: jnp.squeeze(x) if isinstance(x, jax.Array) else x, new_parameters
        )
    new_parameters["metadata"] = metadata

    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        mode="r",
        loads_envelope=sets_envelope,
        loads_metadata=False,
        updates_optics_group=updates_optics_group,
    )
    # Set params
    parameter_file[index] = new_parameters
    # Make sure custom metadata was added
    particle_dataframe = parameter_file.starfile_data["particles"]
    assert set(metadata.columns).issubset(particle_dataframe.columns)
    # Make sure dataframes are the same
    metadata_extracted = particle_dataframe.loc[
        particle_dataframe.index[np.atleast_1d(index)],
        ["rlnMicrographName", "rlnCoordinateX", "rlnCoordinateY"],
    ]
    np.testing.assert_equal(metadata.to_numpy(), metadata_extracted.to_numpy())
    # Load params that were just set
    loaded_parameters = parameter_file[index]
    new_parameters["metadata"] = None
    if updates_optics_group:
        assert compare_pytrees(new_parameters, loaded_parameters)
    else:
        assert compare_pytrees(new_parameters["pose"], loaded_parameters["pose"])
        np.testing.assert_allclose(
            new_parameters["transfer_theory"].ctf.defocus_in_angstroms,  # type: ignore
            loaded_parameters["transfer_theory"].ctf.defocus_in_angstroms,  # type: ignore
        )
        if sets_envelope:
            np.testing.assert_allclose(
                new_parameters["transfer_theory"].envelope.b_factor,  # type: ignore
                loaded_parameters["transfer_theory"].envelope.b_factor,  # type: ignore
            )


def test_file_exists_error():
    # Create pytrees
    parameters = dict(
        image_config=cxs.BasicImageConfig(
            shape=(4, 4), pixel_size=1.1, voltage_in_kilovolts=300.0
        ),
        pose=cxs.EulerAnglePose(),
        transfer_theory=cxs.ContrastTransferTheory(ctf=cxs.AstigmaticCTF()),
    )
    # Add to dataset
    path_to_starfile = "tests/outputs/starfile_writing/test_particle_parameters.star"
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=path_to_starfile,
        mode="w",
        exists_ok=True,
    )
    parameter_file.append(parameters)
    parameter_file.save(overwrite=True)

    # Test no exists_ok
    with pytest.raises(FileExistsError):
        _ = RelionParticleParameterFile(
            path_to_starfile=path_to_starfile,
            mode="w",
            exists_ok=False,
        )
    # Clean up
    shutil.rmtree(parameter_file.path_to_output.parent)


def test_file_not_found_error():
    dummy_path_to_starfile = "path/to/nonexistant/dir/nonexistant_file.star"

    # Test no exists_ok
    with pytest.raises(FileNotFoundError):
        _ = RelionParticleParameterFile(
            path_to_starfile=dummy_path_to_starfile,
            mode="r",
        )

    return


def test_set_wrong_parameters_error():
    # Wrong parameters
    wrong_pose = cxs.QuaternionPose()
    wrong_transfer_theory = cxs.ContrastTransferTheory(
        ctf=cxs.AstigmaticCTF(), envelope=im.FourierDC()
    )
    # Right parameters
    right_pose = cxs.EulerAnglePose()
    right_transfer_theory = cxs.ContrastTransferTheory(ctf=cxs.AstigmaticCTF())
    image_config = cxs.BasicImageConfig(
        shape=(4, 4), pixel_size=1.1, voltage_in_kilovolts=300.0
    )
    # Create pytrees
    wrong_parameters_1 = dict(
        image_config=image_config,
        pose=right_pose,
        transfer_theory=wrong_transfer_theory,
    )
    temp = dict(
        image_config=image_config,
        pose=right_pose,
        transfer_theory=right_transfer_theory,
    )
    wrong_parameters_2 = eqx.tree_at(lambda x: x["pose"], temp, wrong_pose)
    # Now the parameter dataset
    # Add to dataset
    path_to_starfile = "path/to/dummy/project/and/starfile.star"
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=path_to_starfile,
        mode="w",
        exists_ok=True,
    )

    with pytest.raises(ValueError):
        parameter_file.append(wrong_parameters_1)

    with pytest.raises(TypeError):
        parameter_file.append(wrong_parameters_2)


def test_bad_pytree_error():
    # Right parameters
    make_pose = eqx.filter_vmap(
        lambda x, y, phi, theta, psi: cxs.EulerAnglePose(x, y, phi, theta, psi)
    )
    pose = make_pose(
        jnp.atleast_1d(1.0),
        jnp.atleast_1d(-1.0),
        jnp.atleast_1d(1.0),
        jnp.atleast_1d(2.0),
        jnp.atleast_1d(3.0),
    )
    pose = eqx.tree_at(
        lambda x: x.offset_in_angstroms,
        pose,
        jnp.stack(2 * [np.asarray((1.0, 2.0))], axis=-1),
    )
    transfer_theory = cxs.ContrastTransferTheory(ctf=cxs.AstigmaticCTF())
    image_config = cxs.BasicImageConfig(
        shape=(4, 4), pixel_size=1.1, voltage_in_kilovolts=300.0
    )
    # Create pytrees
    parameters = dict(
        image_config=image_config,
        pose=pose,
        transfer_theory=transfer_theory,
    )
    # Now the parameter dataset
    # Add to dataset
    path_to_starfile = "path/to/dummy/project/and/starfile.star"
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=path_to_starfile,
        mode="w",
        exists_ok=True,
    )

    with pytest.raises(ValueError):
        parameter_file.append(parameters)


def test_write_image(
    sample_relion_project_path,
    sample_starfile_path,
    relion_parameters,
):
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        mode="r",
        exists_ok=True,
    )
    parameter_file.path_to_starfile = (
        "tests/outputs/starfile_writing/test_particle_parameters.star"
    )

    dataset = RelionParticleDataset(
        parameter_file,
        path_to_relion_project=sample_relion_project_path,
        mode="w",
    )
    starfile_data = dataset.parameter_file.starfile_data
    assert starfile_data["particles"]["rlnImageName"].isna().all()

    shape = relion_parameters["image_config"].shape
    particle = dict(
        parameters=relion_parameters,
        images=jnp.zeros(shape, dtype=np.float32),
    )
    bad_shape_particle = dict(
        parameters=relion_parameters,
        images=jnp.zeros((shape[0], shape[1] + 1), dtype=np.float32),
    )
    bad_dim_particle = eqx.tree_at(
        lambda x: x["images"], bad_shape_particle, jnp.zeros(shape[0], dtype=np.float32)
    )

    with pytest.raises(ValueError):
        dataset[0] = bad_shape_particle

    with pytest.raises((ValueError, TypeCheckError)):
        dataset[0] = bad_dim_particle

    with pytest.raises(IOError):
        dataset[0] = particle

    dataset.mrcfile_settings = dict(prefix="f", overwrite=True)
    dataset[0] = particle

    starfile_data = dataset.parameter_file.starfile_data
    rln_image_name = starfile_data["particles"]["rlnImageName"][0]
    # Assert entry was written
    assert not pd.isna(rln_image_name)
    assert starfile_data["particles"]["rlnImageName"][1:].isna().all()
    # Assert file was written and delete it
    filename = str(rln_image_name).split("@")[1]
    path_to_filename = os.path.join(sample_relion_project_path, filename)
    assert os.path.exists(path_to_filename)
    os.remove(path_to_filename)
    assert not os.path.exists(path_to_filename)


def test_write_particle_batched_particle_parameters():
    @partial(eqx.filter_vmap, in_axes=(0), out_axes=eqx.if_array(0))
    def _make_particle_params(dummy_idx):
        image_config = cxs.BasicImageConfig(
            shape=(4, 4),
            pixel_size=1.5,
            voltage_in_kilovolts=300.0,
        )

        pose = cxs.EulerAnglePose()
        transfer_theory = cxs.ContrastTransferTheory(
            ctf=cxs.AstigmaticCTF(), envelope=im.FourierGaussian()
        )
        return {
            "image_config": image_config,
            "pose": pose,
            "transfer_theory": transfer_theory,
            "metadata": None,
        }

    particle_params = _make_particle_params(jnp.array([0, 0, 0, 0, 0]))
    new_parameters_file = RelionParticleParameterFile(
        path_to_starfile="dummy.star",
        mode="w",
        exists_ok=True,
        updates_optics_group=True,
        loads_envelope=True,
    )

    new_parameters_file.path_to_starfile = (
        "tests/outputs/starfile_writing/test_particle_parameters.star"
    )

    new_parameters_file.append(particle_params)
    new_parameters_file.save(overwrite=True)
    # and try to save again
    with pytest.raises(FileExistsError):
        new_parameters_file.save(overwrite=False)

    parameter_file = RelionParticleParameterFile(
        path_to_starfile="tests/outputs/starfile_writing/test_particle_parameters.star",
        mode="r",
        loads_envelope=True,
        loads_metadata=False,
    )

    loaded_params = parameter_file[:]
    for key in particle_params:
        assert compare_pytrees(loaded_params[key], particle_params[key]), (
            f"Mismatch for {key}"
        )
    # Clean up
    shutil.rmtree("tests/outputs/starfile_writing/")

    return


def test_write_starfile_different_envs():
    def _make_particle_params(envelope):
        image_config = cxs.BasicImageConfig(
            shape=(4, 4),
            pixel_size=1.5,
            voltage_in_kilovolts=300.0,
        )

        pose = cxs.EulerAnglePose()
        transfer_theory = cxs.ContrastTransferTheory(
            ctf=cxs.AstigmaticCTF(),
            envelope=envelope,
        )
        return {
            "image_config": image_config,
            "pose": pose,
            "transfer_theory": transfer_theory,
            "metadata": None,
        }

    particle_params = _make_particle_params(im.FourierGaussian())
    new_parameters_file = RelionParticleParameterFile(
        path_to_starfile="tests/outputs/starfile_writing/test_particle_parameters.star",
        mode="w",
        exists_ok=True,
        updates_optics_group=True,
        loads_envelope=True,
    )

    particle_params = _make_particle_params(im.FourierConstant(1.0))
    new_parameters_file = RelionParticleParameterFile(
        path_to_starfile="tests/outputs/starfile_writing/test_particle_parameters.star",
        mode="w",
        exists_ok=True,
        updates_optics_group=True,
        loads_envelope=True,
    )
    new_parameters_file.append(particle_params)
    new_parameters_file.save(overwrite=True)

    particle_params = _make_particle_params(None)
    new_parameters_file = RelionParticleParameterFile(
        path_to_starfile="tests/outputs/starfile_writing/test_particle_parameters.star",
        mode="w",
        exists_ok=True,
        updates_optics_group=True,
        loads_envelope=True,
    )
    new_parameters_file.append(particle_params)
    new_parameters_file.save(overwrite=True)

    with pytest.raises(ValueError):
        particle_params = _make_particle_params(im.FourierDC(1.0))
        new_parameters_file = RelionParticleParameterFile(
            path_to_starfile="tests/outputs/starfile_writing/test_particle_parameters.star",
            mode="w",
            exists_ok=True,
            updates_optics_group=True,
            loads_envelope=True,
        )
        new_parameters_file.append(particle_params)
        new_parameters_file.save(overwrite=True)

    # Clean up
    shutil.rmtree("tests/outputs/starfile_writing/")

    return


def test_write_simulated_image_stack_from_starfile_jit(sample_starfile_path):
    def _mock_compute_image(particle_parameters, constant_args, per_particle_args):
        # Mock the image computation
        return per_particle_args

    """Test writing a simulated image stack from a starfile."""
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        mode="r",
        loads_envelope=False,
        loads_metadata=False,
    )
    parameter_file.path_to_starfile = (
        "tests/outputs/starfile_writing/test_particle_parameters.star"
    )

    n_images = len(parameter_file)
    shape = parameter_file[0]["image_config"].shape
    true_images = jax.random.normal(
        jax.random.key(0), shape=(n_images, *shape), dtype=jnp.float32
    )

    # Create a simulated image stack
    new_stack = RelionParticleDataset(
        parameter_file,
        path_to_relion_project="tests/outputs/starfile_writing/",
        mode="w",
        mrcfile_settings={"overwrite": True},
    )

    simulate_particle_stack(
        new_stack,
        simulate_fn=_mock_compute_image,
        constant_args=(1.0, 2.0),
        per_particle_args=true_images,
        is_jittable=True,
        overwrite=True,
    )

    # try to overwrite
    simulate_particle_stack(
        new_stack,
        simulate_fn=_mock_compute_image,
        constant_args=(1.0, 2.0),
        per_particle_args=true_images,
        is_jittable=True,
        overwrite=True,
    )

    # Now trigger overwrite error
    with pytest.raises(FileExistsError):
        simulate_particle_stack(
            new_stack,
            simulate_fn=_mock_compute_image,
            constant_args=(1.0, 2.0),
            per_particle_args=true_images,
            is_jittable=True,
            overwrite=False,
        )

    # load the simulated image stack
    particle_dataset = RelionParticleDataset(
        parameter_file,
        path_to_relion_project="tests/outputs/starfile_writing/",
        mode="r",
    )

    images = particle_dataset[:]["images"]
    np.testing.assert_allclose(
        images,
        true_images.astype(jnp.float32),
    )

    # Clean up
    shutil.rmtree("tests/outputs/starfile_writing/")

    return


def test_write_simulated_image_stack_from_starfile_nojit(sample_starfile_path):
    def _mock_compute_image(particle_parameters, constant_args, per_particle_args):
        # Mock the image computation
        c1, c2 = constant_args
        image = per_particle_args
        return image / jnp.linalg.norm(image)

    """Test writing a simulated image stack from a starfile."""
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        mode="r",
        loads_envelope=False,
        loads_metadata=False,
    )
    parameter_file.path_to_starfile = (
        "tests/outputs/starfile_writing/test_particle_parameters.star"
    )

    n_images = len(parameter_file)
    shape = parameter_file[0]["image_config"].shape
    true_images = jax.random.normal(
        jax.random.key(0), shape=(n_images, *shape), dtype=jnp.float32
    )

    # Create a simulated image stack
    new_stack = RelionParticleDataset(
        parameter_file,
        path_to_relion_project="tests/outputs/starfile_writing/",
        mode="w",
        mrcfile_settings={"overwrite": True},
    )

    simulate_particle_stack(
        new_stack,
        simulate_fn=_mock_compute_image,
        constant_args=(1.0, 2.0),
        per_particle_args=true_images,
        overwrite=True,
    )

    particle_dataset = RelionParticleDataset(
        parameter_file,
        path_to_relion_project="tests/outputs/starfile_writing/",
        mode="r",
    )
    images = particle_dataset[:]["images"]
    assert jnp.allclose(
        images,
        true_images / np.linalg.norm(true_images, axis=(1, 2), keepdims=True),
    )

    # Clean up
    shutil.rmtree("tests/outputs/starfile_writing/")

    return


def test_write_single_image(sample_starfile_path):
    def _mock_compute_image(particle_parameters, constant_args, per_particle_args):
        # Mock the image computation
        c1, c2 = constant_args
        p1, p2 = per_particle_args
        image = jnp.ones(particle_parameters["image_config"].shape, dtype=jnp.float32)
        return image / jnp.linalg.norm(image)

    selection_filter = {
        "rlnImageName": lambda x: np.where(x == "0000001@000000.mrcs", True, False)
    }
    """Test writing a simulated image stack from a starfile."""
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        mode="r",
        loads_envelope=False,
        loads_metadata=False,
        selection_filter=selection_filter,
    )
    parameter_file.path_to_starfile = (
        "tests/outputs/starfile_writing/test_particle_parameters.star"
    )

    # Create a simulated image stack
    new_stack = RelionParticleDataset(
        parameter_file,
        path_to_relion_project="tests/outputs/starfile_writing/",
        mode="w",
        mrcfile_settings={"overwrite": True},
    )

    n_images = 1
    simulate_particle_stack(
        new_stack,
        simulate_fn=_mock_compute_image,
        constant_args=(1.0, 2.0),
        per_particle_args=(3.0 * jnp.ones(n_images), 4.0 * jnp.ones(n_images)),
        overwrite=True,
        images_per_file=1,
    )

    particle_dataset = RelionParticleDataset(
        parameter_file,
        path_to_relion_project="tests/outputs/starfile_writing/",
        mode="r",
    )
    images = particle_dataset[:]["images"]
    np.testing.assert_allclose(
        images,
        np.ones_like(images) / np.linalg.norm(np.ones_like(images)),
    )

    # Clean up
    shutil.rmtree("tests/outputs/starfile_writing/")

    return


def test_load_multiple_mrcs():
    @partial(eqx.filter_vmap, in_axes=(0), out_axes=eqx.if_array(0))
    def _make_particle_params(dummy_idx):
        image_config = cxs.BasicImageConfig(
            shape=(4, 4),
            pixel_size=1.5,
            voltage_in_kilovolts=300.0,
        )

        pose = cxs.EulerAnglePose()
        transfer_theory = cxs.ContrastTransferTheory(
            ctf=cxs.AstigmaticCTF(), envelope=im.FourierGaussian()
        )
        return {
            "image_config": image_config,
            "pose": pose,
            "transfer_theory": transfer_theory,
        }

    def _mock_compute_image(particle_parameters, constant_args, per_particle_args):
        # Mock the image computation
        return per_particle_args

    particle_params = _make_particle_params(jnp.ones(10))
    parameters_file = RelionParticleParameterFile(
        path_to_starfile="tests/outputs/starfile_writing/test_particle_parameters.star",
        mode="w",
        exists_ok=True,
        updates_optics_group=True,
        loads_envelope=True,
    )
    parameters_file.append(particle_params)

    n_images = len(parameters_file)
    # print(f"Number of images: {n_images}")
    shape = parameters_file[0]["image_config"].shape
    true_images = jax.random.normal(
        jax.random.key(0), shape=(n_images, *shape), dtype=jnp.float32
    )

    new_dataset = RelionParticleDataset(
        parameters_file,
        path_to_relion_project="tests/outputs/starfile_writing/",
        mode="w",
        mrcfile_settings={"overwrite": True},
    )

    # Create a simulated image stack
    simulate_particle_stack(
        new_dataset,
        simulate_fn=_mock_compute_image,
        constant_args=(1.0, 2.0),
        per_particle_args=true_images,
        overwrite=True,
        images_per_file=3,
        batch_size=3,
    )

    n_tests = 10
    for _ in range(n_tests):
        indices = np.random.choice(len(parameters_file), size=3, replace=False)

        images = new_dataset[indices]["images"]
        np.testing.assert_allclose(
            images,
            true_images[indices],
        )

    shutil.rmtree("tests/outputs/starfile_writing/")
    return


def test_raise_errors_parameter_file(sample_starfile_path):
    from jaxtyping import TypeCheckError

    with pytest.raises((ValueError, TypeCheckError)):
        parameter_file = RelionParticleParameterFile(
            path_to_starfile=sample_starfile_path,
            mode="CRYOEM",  # type: ignore
            loads_envelope=False,
            loads_metadata=False,
        )
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        mode="r",
        loads_envelope=False,
        loads_metadata=False,
    )

    assert parameter_file.mode == "r"

    # bad keys, bad values
    with pytest.raises((ValueError, TypeCheckError)):
        starfile_data = {"cryo": pd.DataFrame({}), "em": pd.DataFrame({})}
        parameter_file.starfile_data = starfile_data

    # good keys, bad values
    with pytest.raises((ValueError, TypeCheckError)):
        starfile_data = {"particles": 0, "optics": 0}
        parameter_file.starfile_data = starfile_data  # type: ignore

    # now set to write mode and try to filter
    with pytest.raises(ValueError):
        parameter_file = RelionParticleParameterFile(
            path_to_starfile=sample_starfile_path,
            mode="w",
            exists_ok=True,
            loads_envelope=False,
            loads_metadata=False,
            selection_filter={"rlnAngleRot": lambda x: x < 1000.0},
        )


def test_raise_errors_stack_dataset(sample_starfile_path, sample_relion_project_path):
    parameter_file = RelionParticleParameterFile(
        path_to_starfile=sample_starfile_path,
        mode="r",
        loads_envelope=False,
        loads_metadata=False,
    )
    parameter_file.path_to_starfile = (
        "tests/outputs/starfile_writing/test_particle_parameters.star"
    )

    starfile_data = parameter_file.starfile_data
    particle_data = starfile_data["particles"]
    # remove "rlnImageName" column
    particle_data = particle_data.drop(columns=["rlnImageName"])

    parameter_file.starfile_data = {
        "particles": particle_data,
        "optics": starfile_data["optics"],
    }

    with pytest.raises(IOError):
        particle_dataset = RelionParticleDataset(
            parameter_file,
            path_to_relion_project=sample_relion_project_path,
            mode="r",
        )

    # Now set to write mode

    parameter_file.path_to_starfile = (
        "tests/outputs/starfile_writing/test_particle_parameters.star"
    )
    particle_dataset = RelionParticleDataset(
        parameter_file,
        path_to_relion_project="tests/outputs/starfile_writing/",
        mode="w",
        mrcfile_settings={"overwrite": False},
    )

    parameters = parameter_file[0]
    image_shape = parameters["image_config"].shape

    particle_stack = {
        "parameters": parameters,
        "images": jnp.zeros(image_shape, dtype=np.float32),
    }

    with pytest.raises(ValueError):
        particle_dataset[np.array([0])] = particle_stack

    with pytest.raises(TypeError):
        particle_dataset[0] = "dummy"  # type: ignore

    with pytest.raises(TypeError):
        particle_dataset.append("dummy")  # type: ignore

    with pytest.raises(ValueError):
        particle_dataset.append({"parameters": None, "images": particle_stack["images"]})

    with pytest.raises(ValueError):
        particle_dataset.write_images(
            index_array=np.array([0, 1], dtype=int), images=np.zeros((100, 10, 10))
        )

    with pytest.raises(ValueError):
        particle_dataset.write_images(
            index_array=np.array([0, 1], dtype=int), images=np.zeros((10, *image_shape))
        )

    # and clean
    shutil.rmtree("tests/outputs/starfile_writing/")


def test_append_relion_stack_dataset():
    @partial(eqx.filter_vmap, in_axes=(0), out_axes=eqx.if_array(0))
    def _make_particle_params(dummy_idx):
        image_config = cxs.BasicImageConfig(
            shape=(4, 4),
            pixel_size=1.5,
            voltage_in_kilovolts=300.0,
        )

        pose = cxs.EulerAnglePose()
        transfer_theory = cxs.ContrastTransferTheory(
            ctf=cxs.AstigmaticCTF(), envelope=im.FourierGaussian()
        )
        return {
            "image_config": image_config,
            "pose": pose,
            "transfer_theory": transfer_theory,
            "metadata": None,
        }

    def _mock_compute_image(particle_parameters, constant_args, per_particle_args):
        # Mock the image computation
        return per_particle_args

    new_parameters_file = RelionParticleParameterFile(
        path_to_starfile="tests/outputs/starfile_writing/test_particle_parameters.star",
        mode="w",
        exists_ok=True,
        updates_optics_group=True,
        loads_envelope=True,
    )

    new_stack = RelionParticleDataset(
        new_parameters_file,
        path_to_relion_project="tests/outputs/starfile_writing/",
        mode="w",
        mrcfile_settings={"overwrite": False},
    )

    n_images = 10
    particle_params = _make_particle_params(jnp.ones(n_images))
    shape = particle_params["image_config"].shape
    images = jax.random.normal(
        jax.random.key(0), shape=(n_images, *shape), dtype=jnp.float32
    )

    new_stack.append(
        {
            "parameters": particle_params,
            "images": images,
        }
    )

    # clean up
    shutil.rmtree("tests/outputs/starfile_writing/")
    return



import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, Float

from ..jax_util import NDArrayLike


def b_factor_to_variance(
    b_factor: Float[NDArrayLike, "..."] | float,
) -> Float[Array, "..."]:
    """From the B-factor of a gaussian, return the corresponding
    value of the variance of the gaussian. This simply does the
    conversion

    $$\\sigma^2 = B / (8 \\pi^2)$$

    The purpose of this function is to make it easy to convert
    between conventions when defining gaussians.

    **Arguments:**

    - `b_factor`:
        The B-factor of the gaussian.

    **Returns:**

    The variance.
    """
    return jnp.asarray(b_factor / (8 * np.pi**2))


def variance_to_b_factor(
    variance: Float[NDArrayLike, "..."] | float,
) -> Float[Array, "..."]:
    """From the variance of a gaussian, return the corresponding
    value of the B-factor of the gaussian. This simply does the
    conversion

    $$B = 8 \\pi^2 * \\sigma^2$$

    The purpose of this function is to make it easy to convert
    between conventions when defining gaussians.

    **Arguments:**

    - `variance`:
        The variance of the gaussian.

    **Returns:**

    The B-factor.
    """
    return jnp.asarray((8 * np.pi**2) * variance)
